from coredis.exceptions import DataError, RedisClusterException, RedisError
from coredis.utils import b, bool_ok, dict_merge, nativestr, string_keys_to_dict


class ListsCommandMixin:

    RESPONSE_CALLBACKS = dict_merge(
        string_keys_to_dict("BLPOP BRPOP", lambda r: r and tuple(r) or None),
        string_keys_to_dict(
            # these return OK, or int if redis-server is >=1.3.4
            "LPUSH RPUSH",
            lambda r: isinstance(r, int) and r or nativestr(r) == "OK",
        ),
        string_keys_to_dict("LSET LTRIM", bool_ok),
        string_keys_to_dict("LINSERT LLEN LPUSHX RPUSHX", int),
    )

    async def blpop(self, keys, timeout=0):
        """
        LPOP a value off of the first non-empty list
        named in the ``keys`` list.

        If none of the lists in ``keys`` has a value to LPOP, then block
        for ``timeout`` seconds, or until a value gets pushed on to one
        of the lists.

        If timeout is 0, then block indefinitely.
        """

        if timeout is None:
            timeout = 0

        if isinstance(keys, str):
            keys = [keys]
        else:
            keys = list(keys)
        keys.append(timeout)

        return await self.execute_command("BLPOP", *keys)

    async def brpop(self, keys, timeout=0):
        """
        RPOP a value off of the first non-empty list
        named in the ``keys`` list.

        If none of the lists in ``keys`` has a value to LPOP, then block
        for ``timeout`` seconds, or until a value gets pushed on to one
        of the lists.

        If timeout is 0, then block indefinitely.
        """

        if timeout is None:
            timeout = 0

        if isinstance(keys, str):
            keys = [keys]
        else:
            keys = list(keys)
        keys.append(timeout)

        return await self.execute_command("BRPOP", *keys)

    async def brpoplpush(self, src, dst, timeout=0):
        """
        Pop a value off the tail of ``src``, push it on the head of ``dst``
        and then return it.

        This command blocks until a value is in ``src`` or until ``timeout``
        seconds elapse, whichever is first. A ``timeout`` value of 0 blocks
        forever.
        """

        if timeout is None:
            timeout = 0

        return await self.execute_command("BRPOPLPUSH", src, dst, timeout)

    async def lindex(self, name, index):
        """
        Return the item from list ``name`` at position ``index``

        Negative indexes are supported and will return an item at the
        end of the list
        """

        return await self.execute_command("LINDEX", name, index)

    async def linsert(self, name, where, refvalue, value):
        """
        Insert ``value`` in list ``name`` either immediately before or after
        [``where``] ``refvalue``

        Returns the new length of the list on success or -1 if ``refvalue``
        is not in the list.
        """

        return await self.execute_command("LINSERT", name, where, refvalue, value)

    async def llen(self, name):
        """Returns the length of the list ``name``"""

        return await self.execute_command("LLEN", name)

    async def lpop(self, name):
        """RemoveS and returns the first item of the list ``name``"""

        return await self.execute_command("LPOP", name)

    async def lpush(self, name, *values):
        """Pushes ``values`` onto the head of the list ``name``"""

        return await self.execute_command("LPUSH", name, *values)

    async def lpushx(self, name, value):
        """
        Pushes ``value`` onto the head of the list ``name`` if ``name`` exists
        """

        return await self.execute_command("LPUSHX", name, value)

    async def lrange(self, name, start, end):
        """
        Returns a slice of the list ``name`` between
        position ``start`` and ``end``

        ``start`` and ``end`` can be negative numbers just like
        Python slicing notation
        """

        return await self.execute_command("LRANGE", name, start, end)

    async def lrem(self, name, count, value):
        """
        Removes the first ``count`` occurrences of elements equal to ``value``
        from the list stored at ``name``.

        The count argument influences the operation in the following ways:
            count > 0: Remove elements equal to value moving from head to tail.
            count < 0: Remove elements equal to value moving from tail to head.
            count = 0: Remove all elements equal to value.
        """

        return await self.execute_command("LREM", name, count, value)

    async def lset(self, name, index, value):
        """Sets ``position`` of list ``name`` to ``value``"""

        return await self.execute_command("LSET", name, index, value)

    async def ltrim(self, name, start, end):
        """
        Trims the list ``name``, removing all values not within the slice
        between ``start`` and ``end``

        ``start`` and ``end`` can be negative numbers just like
        Python slicing notation
        """

        return await self.execute_command("LTRIM", name, start, end)

    async def rpop(self, name):
        """Removes and return the last item of the list ``name``"""

        return await self.execute_command("RPOP", name)

    async def rpoplpush(self, src, dst):
        """
        RPOP a value off of the ``src`` list and atomically LPUSH it
        on to the ``dst`` list.  Returns the value.
        """

        return await self.execute_command("RPOPLPUSH", src, dst)

    async def rpush(self, name, *values):
        """Pushes ``values`` onto the tail of the list ``name``"""

        return await self.execute_command("RPUSH", name, *values)

    async def rpushx(self, name, value):
        """
        Pushes ``value`` onto the tail of the list ``name`` if ``name`` exists
        """

        return await self.execute_command("RPUSHX", name, value)

    async def lmove(self, first_list, second_list, src="LEFT", dest="RIGHT"):
        """
        Atomically returns and removes the first/last element of a list,
        pushing it as the first/last element on the destination list.
        Returns the element being popped and pushed.

        .. versionadded:: 2.1.0
        """
        params = [first_list, second_list, src, dest]

        return await self.execute_command("LMOVE", *params)

    async def blmove(self, first_list, second_list, timeout, src="LEFT", dest="RIGHT"):
        """
        Blocking version of lmove.

        .. versionadded:: 2.1.0
        """
        params = [first_list, second_list, src, dest, timeout]

        return await self.execute_command("BLMOVE", *params)

    async def lpos(self, name, value, rank=None, count=None, maxlen=None):
        """
        Get position of ``value`` within the list ``name``
         If specified, ``rank`` indicates the "rank" of the first element to
         return in case there are multiple copies of ``value`` in the list.
         By default, LPOS returns the position of the first occurrence of
         ``value`` in the list. When ``rank`` 2, LPOS returns the position of
         the second ``value`` in the list. If ``rank`` is negative, LPOS
         searches the list in reverse. For example, -1 would return the
         position of the last occurrence of ``value`` and -2 would return the
         position of the next to last occurrence of ``value``.
         If specified, ``count`` indicates that LPOS should return a list of
         up to ``count`` positions. A ``count`` of 2 would return a list of
         up to 2 positions. A ``count`` of 0 returns a list of all positions
         matching ``value``. When ``count`` is specified and but ``value``
         does not exist in the list, an empty list is returned.
         If specified, ``maxlen`` indicates the maximum number of list
         elements to scan. A ``maxlen`` of 1000 will only return the
         position(s) of items within the first 1000 entries in the list.
         A ``maxlen`` of 0 (the default) will scan the entire list.

        .. versionadded:: 2.1.0
        """
        pieces = [name, value]

        if rank is not None:
            pieces.extend(["RANK", rank])

        if count is not None:
            pieces.extend(["COUNT", count])

        if maxlen is not None:
            pieces.extend(["MAXLEN", maxlen])

        return await self.execute_command("LPOS", *pieces)


class ClusterListsCommandMixin(ListsCommandMixin):
    async def brpoplpush(self, src, dst, timeout=0):
        """
        Pops a value off the tail of ``src``, push it on the head of ``dst``
        and then return it.

        This command blocks until a value is in ``src`` or until ``timeout``
        seconds elapse, whichever is first. A ``timeout`` value of 0 blocks
        forever.

        Cluster impl:
            Call brpop() then send the result into lpush()

            Operation is no longer atomic.
        """
        try:
            value = await self.brpop(src, timeout=timeout)

            if value is None:
                return None
        except TimeoutError:
            # Timeout was reached

            return None

        await self.lpush(dst, value[1])

        return value[1]

    async def rpoplpush(self, src, dst):
        """
        RPOP a value off of the ``src`` list and atomically LPUSH it
        on to the ``dst`` list.  Returns the value.

        Cluster impl:
            Call rpop() then send the result into lpush()

            Operation is no longer atomic.
        """
        value = await self.rpop(src)

        if value:
            await self.lpush(dst, value)

            return value

        return None

    async def sort(
        self,
        name,
        start=None,
        num=None,
        by=None,
        get=None,
        desc=False,
        alpha=False,
        store=None,
        groups=None,
    ):
        """Sorts and returns a list, set or sorted set at ``name``.

        :start: and :num:
            are for paging through the sorted data

        :by:
            allows using an external key to weight and sort the items.
            Use an "*" to indicate where in the key the item value is located

        :get:
            is for returning items from external keys rather than the
            sorted data itself.  Use an "*" to indicate where int he key
            the item value is located

        :desc:
            is for reversing the sort

        :alpha:
            is for sorting lexicographically rather than numerically

        :store:
            is for storing the result of the sort into the key `store`

        ClusterImpl:
            A full implementation of the server side sort mechanics because many of the
            options work on multiple keys that can exist on multiple servers.
        """

        if (start is None and num is not None) or (start is not None and num is None):
            raise RedisError("RedisError: ``start`` and ``num`` must both be specified")
        try:
            data_type = b(await self.type(name))

            if data_type == b("none"):
                return []
            elif data_type == b("set"):
                data = list(await self.smembers(name))[:]
            elif data_type == b("list"):
                data = await self.lrange(name, 0, -1)
            else:
                raise RedisClusterException(
                    "Unable to sort data type : {0}".format(data_type)
                )

            if by is not None:
                # _sort_using_by_arg mutates data so we don't
                # need need a return value.
                data = await self._sort_using_by_arg(data, by, alpha)
            elif not alpha:
                data.sort(key=self._strtod_key_func)
            else:
                data.sort()

            if desc:
                data = data[::-1]

            if not (start is None and num is None):
                data = data[start : start + num]

            if get:
                data = await self._retrive_data_from_sort(data, get)

            if store is not None:
                if data_type == b("set"):
                    await self.delete(store)
                    await self.rpush(store, *data)
                elif data_type == b("list"):
                    await self.delete(store)
                    await self.rpush(store, *data)
                else:
                    raise RedisClusterException(
                        "Unable to store sorted data for data type : {0}".format(
                            data_type
                        )
                    )

                return len(data)

            if groups:
                if not get or isinstance(get, str) or len(get) < 2:
                    raise DataError(
                        'when using "groups" the "get" argument '
                        "must be specified and contain at least "
                        "two keys"
                    )
                n = len(get)

                return list(zip(*[data[i::n] for i in range(n)]))
            else:
                return data
        except KeyError:
            return []

    async def _retrive_data_from_sort(self, data, get):
        """
        Used by sort()
        """

        if get is not None:
            if isinstance(get, str):
                get = [get]
            new_data = []

            for k in data:
                for g in get:
                    single_item = await self._get_single_item(k, g)
                    new_data.append(single_item)
            data = new_data

        return data

    async def _get_single_item(self, k, g):
        """
        Used by sort()
        """

        if getattr(k, "decode", None):
            k = k.decode("utf-8")

        if "*" in g:
            g = g.replace("*", k)

            if "->" in g:
                key, hash_key = g.split("->")
                single_item = await self.get(key, {}).get(hash_key)
            else:
                single_item = await self.get(g)
        elif "#" in g:
            single_item = k
        else:
            single_item = None

        return b(single_item)

    def _strtod_key_func(self, arg):
        """
        Used by sort()
        """

        return float(arg)

    async def _sort_using_by_arg(self, data, by, alpha):
        """
        Used by sort()
        """

        if getattr(by, "decode", None):
            by = by.decode("utf-8")

        async def _by_key(arg):
            if getattr(arg, "decode", None):
                arg = arg.decode("utf-8")

            key = by.replace("*", arg)

            if "->" in by:
                key, hash_key = key.split("->")
                v = await self.hget(key, hash_key)

                if alpha:
                    return v
                else:
                    return float(v)
            else:
                return await self.get(key)

        sorted_data = []

        for d in data:
            sorted_data.append((d, await _by_key(d)))

        return [x[0] for x in sorted(sorted_data, key=lambda x: x[1])]



from coredis.exceptions import RedisError
from coredis.utils import bool_ok, dict_merge, pairs_to_dict, string_keys_to_dict


def stream_list(response):
    result = []
    if response:
        for r in response:
            kv_pairs = r[1]
            kv_dict = dict()
            while kv_pairs and len(kv_pairs) > 1:
                kv_dict[kv_pairs.pop()] = kv_pairs.pop()
            result.append((r[0], kv_dict))
    return result


def multi_stream_list(response):
    result = dict()
    if response:
        for r in response:
            result[r[0]] = stream_list(r[1])
    return result


def list_of_pairs_to_dict(response):
    return [pairs_to_dict(row) for row in response]


def parse_xinfo_stream(response):
    res = pairs_to_dict(response)
    if res["first-entry"] and len(res["first-entry"]) > 0:
        res["first-entry"][1] = pairs_to_dict(res["first-entry"][1])
    if res["last-entry"] and len(res["last-entry"]) > 0:
        res["last-entry"][1] = pairs_to_dict(res["last-entry"][1])
    return res


class StreamsCommandMixin:
    RESPONSE_CALLBACKS = dict_merge(
        string_keys_to_dict("XREVRANGE XRANGE", stream_list),
        string_keys_to_dict("XREAD XREADGROUP", multi_stream_list),
        {
            "XINFO GROUPS": list_of_pairs_to_dict,
            "XINFO STREAM": parse_xinfo_stream,
            "XINFO CONSUMERS": list_of_pairs_to_dict,
            "XGROUP SETID": bool_ok,
            "XGROUP CREATE": bool_ok,
        },
    )

    async def xadd(
        self, name: str, entry: dict, max_len=None, stream_id="*", approximate=True
    ) -> str:
        """
        Appends the specified stream entry to the stream at the specified key.
        If the key does not exist, as a side effect of running
        this command the key is created with a stream value.

        :param name: name of the stream
        :param entry: key-values to be appended to the stream
        :param max_len: max length of the stream
         length will not be limited max_len is set to None
         notice: max_len should be int greater than 0,
         if set to 0 or negative, the stream length will not be limited
        :param stream_id: id of the options appended to the stream.
         The XADD command will auto-generate a unique id for you
         if the id argument specified is the * character.
         ID are specified by two numbers separated by a "-" character
        :param approximate: whether redis will limit
         the stream with given max length exactly, if set to True,
         there will be a few tens of entries more,
         but never less than 1000 items
        :return: id auto generated or the specified id given.

        .. note:: specified id without "-" character will be completed like "id-0"
        """
        pieces = []
        if max_len is not None:
            if not isinstance(max_len, int) or max_len < 1:
                raise RedisError("XADD maxlen must be a positive integer")
            pieces.append("MAXLEN")
            if approximate:
                pieces.append("~")
            pieces.append(str(max_len))
        pieces.append(stream_id)
        for kv in entry.items():
            pieces.extend(list(kv))
        return await self.execute_command("XADD", name, *pieces)

    async def xlen(self, name: str) -> int:
        """
        Returns the number of elements in a given stream.
        """
        return await self.execute_command("XLEN", name)

    async def xrange(self, name: str, start="-", end="+", count=None) -> list:
        """
        Read stream values within an interval.

        :param name: name of the stream.
        :param start: first stream ID. defaults to '-',
         meaning the earliest available.
        :param end: last stream ID. defaults to '+',
         meaning the latest available.
        :param count: if set, only return this many items, beginning with the
         earliest available.
        :return: list of (stream_id, entry(k-v pair))
        """

        pieces = [start, end]
        if count is not None:
            if not isinstance(count, int) or count < 1:
                raise RedisError("XRANGE count must be a positive integer")
            pieces.append("COUNT")
            pieces.append(str(count))
        return await self.execute_command("XRANGE", name, *pieces)

    async def xrevrange(self, name: str, start="+", end="-", count=None) -> list:
        """
        Read stream values within an interval, in reverse order.

        :param name: name of the stream
        :param start: first stream ID. defaults to '+',
               meaning the latest available.
        :param end: last stream ID. defaults to '-',
                meaning the earliest available.
        :param count: if set, only return this many items, beginning with the
               latest available.

        """
        pieces = [start, end]
        if count is not None:
            if not isinstance(count, int) or count < 1:
                raise RedisError("XREVRANGE count must be a positive integer")
            pieces.append("COUNT")
            pieces.append(str(count))
        return await self.execute_command("XREVRANGE", name, *pieces)

    async def xread(self, count=None, block=None, **streams) -> dict:
        """
        Read data from one or multiple streams,
        only returning entries with an ID greater
        than the last received ID reported by the caller.

        :param count: int, if set, only return this many items, beginning with the
               earliest available.
        :param block: int, milliseconds we want to block before timing out,
                if the BLOCK option is not used, the command is synchronous
        :param streams: stream_name - stream_id mapping
        :return dict like {stream_name: [(stream_id: entry), ...]}
        """
        pieces = []
        if block is not None:
            if not isinstance(block, int) or block < 0:
                raise RedisError("XREAD block must be a positive integer")
            pieces.append("BLOCK")
            pieces.append(str(block))
        if count is not None:
            if not isinstance(count, int) or count < 1:
                raise RedisError("XREAD count must be a positive integer")
            pieces.append("COUNT")
            pieces.append(str(count))
        pieces.append("STREAMS")
        ids = []
        for partial_stream in streams.items():
            pieces.append(partial_stream[0])
            ids.append(partial_stream[1])
        pieces.extend(ids)
        return await self.execute_command("XREAD", *pieces)

    async def xreadgroup(
        self, group: str, consumer_id: str, count=None, block=None, **streams
    ):
        """
        Read data from one or multiple streams via the consumer group,
        only returning entries with an ID greater
        than the last received ID reported by the caller.

        :param group: the name of the consumer group
        :param consumer_id: the name of the consumer that is attempting to read
        :param count: int, if set, only return this many items, beginning with the
               earliest available.
        :param block: int, milliseconds we want to block before timing out,
                if the BLOCK option is not used, the command is synchronous
        :param streams: stream_name - stream_id mapping
        :return dict like {stream_name: [(stream_id: entry), ...]}
        """
        pieces = ["GROUP", group, consumer_id]
        if block is not None:
            if not isinstance(block, int) or block < 1:
                raise RedisError("XREAD block must be a positive integer")
            pieces.append("BLOCK")
            pieces.append(str(block))
        if count is not None:
            if not isinstance(count, int) or count < 1:
                raise RedisError("XREAD count must be a positive integer")
            pieces.append("COUNT")
            pieces.append(str(count))
        pieces.append("STREAMS")
        ids = []
        for partial_stream in streams.items():
            pieces.append(partial_stream[0])
            ids.append(partial_stream[1])
        pieces.extend(ids)
        return await self.execute_command("XREADGROUP", *pieces)

    async def xpending(
        self, name: str, group: str, start="-", end="+", count=None, consumer=None
    ) -> list:
        """
        Fetching data from a stream via a consumer group,
        and not acknowledging such data,
        has the effect of creating pending entries.
        The XPENDING command is the interface to inspect the list of pending messages.

        :param name: name of the stream
        :param group: name of the consumer group
        :param start: first stream ID. defaults to '-',
               meaning the earliest available.
        :param end: last stream ID. defaults to '+',
                meaning the latest available.
        :param count: int, number of entries
                [NOTICE] only when count is set to int,
                start & end options will have effect
                and detail of pending entries will be returned
        :param consumer: str, consumer of the stream in the group
                [NOTICE] only when count is set to int,
                this option can be appended to
                query pending entries of given consumer
        """
        pieces = [name, group]
        if count is not None:
            pieces.extend([start, end, count])
            if consumer is not None:
                pieces.append(str(consumer))
        # todo: may there be a parse function
        return await self.execute_command("XPENDING", *pieces)

    async def xtrim(self, name: str, max_len: int, approximate=True) -> int:
        """
        XTRIM is designed to accept different trimming strategies,
        even if currently only MAXLEN is implemented.

        :param name: name of the stream
        :param max_len: max length of the stream after being trimmed
        :param approximate: whether redis will limit
         the stream with given max length exactly, if set to True,
         there will be a few tens of entries more,
         but never less than 1000 items:

        :return: number of entries trimmed
        """
        pieces = ["MAXLEN"]
        if approximate:
            pieces.append("~")
        pieces.append(max_len)
        return await self.execute_command("XTRIM", name, *pieces)

    async def xdel(self, name: str, stream_id: str) -> int:
        """
        [NOTICE] In the current implementation, memory is not
        really reclaimed until a macro node is completely empty,
        so you should not abuse this feature.

        remove items from the middle of a stream, just by ID.

        :param name: name of the stream
        :param stream_id: id of the options appended to the stream.
        """
        return await self.execute_command("XDEL", name, stream_id)

    async def xinfo_consumers(self, name: str, group: str) -> list:
        """
        XINFO command is an observability interface that can be used
        with sub-commands in order to get information
        about streams or consumer groups.

        :param name: name of the stream
        :param group: name of the consumer group
        """
        return await self.execute_command("XINFO CONSUMERS", name, group)

    async def xinfo_groups(self, name: str) -> list:
        """
        XINFO command is an observability interface that can be used
        with sub-commands in order to get information
        about streams or consumer groups.

        :param name: name of the stream
        """
        return await self.execute_command("XINFO GROUPS", name)

    async def xinfo_stream(self, name: str) -> dict:
        """
        XINFO command is an observability interface that can be used
        with sub-commands in order to get information
        about streams or consumer groups.

        :param name: name of the stream
        """
        return await self.execute_command("XINFO STREAM", name)

    async def xack(self, name: str, group: str, stream_id: str) -> int:
        """
        XACK is the command that allows a consumer to mark a pending message as correctly processed.

        :param name: name of the stream
        :param group: name of the consumer group
        :param stream_id: id of the entry the consumer wants to mark
        :return: number of entry marked
        """
        return await self.execute_command("XACK", name, group, stream_id)

    async def xclaim(
        self, name: str, group: str, consumer: str, min_idle_time: int, *stream_ids
    ):
        """
        Gets ownership of one or multiple messages in the Pending Entries List of a given stream
        consumer group.

        :param name: name of the stream
        :param group: name of the consumer group
        :param consumer: name of the consumer
        :param min_idle_time: ms
            If the message ID (among the specified ones) exists, and its idle time greater
            or equal to min_idle_time, then the message new owner
            becomes the specified <consumer>. If the minimum idle time specified
            is zero, messages are claimed regardless of their idle time.
        :param stream_ids:
        """
        return await self.execute_command(
            "XCLAIM", name, group, consumer, min_idle_time, *stream_ids
        )

    async def xgroup_create(self, name: str, group: str, stream_id="$") -> bool:
        """
        XGROUP is used in order to create, destroy and manage consumer groups.

        :param name: name of the stream
        :param group: name of the consumer group
        :param stream_id:
            If we provide $ as we did, then only new messages arriving
            in the stream from now on will be provided to the consumers in the group.
            If we specify 0 instead the consumer group will consume all the messages
            in the stream history to start with.
            Of course, you can specify any other valid ID
        """
        return await self.execute_command("XGROUP CREATE", name, group, stream_id)

    async def xgroup_set_id(self, name: str, group: str, stream_id: str) -> bool:
        """
        :param name: name of the stream
        :param group: name of the consumer group
        :param stream_id:
            If we provide $ as we did, then only new messages arriving
            in the stream from now on will be provided to the consumers in the group.
            If we specify 0 instead the consumer group will consume all the messages
            in the stream history to start with.
            Of course, you can specify any other valid ID
        """
        return await self.execute_command("XGROUP SETID", name, group, stream_id)

    async def xgroup_destroy(self, name: str, group: str) -> int:
        """
        XGROUP is used in order to create, destroy and manage consumer groups.

        :param name: name of the stream
        :param group: name of the consumer group
        """
        return await self.execute_command("XGROUP DESTROY", name, group)

    async def xgroup_del_consumer(self, name: str, group: str, consumer: str) -> int:
        """
        XGROUP is used in order to create, destroy and manage consumer groups.

        :param name: name of the stream
        :param group: name of the consumer group
        :param consumer: name of the consumer
        """
        return await self.execute_command("XGROUP DELCONSUMER", name, group, consumer)



import datetime
import time

from coredis.exceptions import DataError, ReadOnlyError, RedisError
from coredis.utils import (
    NodeFlag,
    bool_ok,
    dict_merge,
    iteritems,
    list_or_args,
    nativestr,
    string_keys_to_dict,
)


class BitFieldOperation:
    """
    The command treats a Redis string as a array of bits,
    and is capable of addressing specific integer fields
    of varying bit widths and arbitrary non (necessary) aligned offset.

    The supported types are up to 64 bits for signed integers,
    and up to 63 bits for unsigned integers.

    Offset can be num prefixed with `#` character or num directly,
    for command detail you should see: https://redis.io/commands/bitfield
    """

    def __init__(self, redis_client, key, readonly=False):
        self._command_stack = ["BITFIELD" if not readonly else "BITFIELD_RO", key]
        self.redis = redis_client
        self.readonly = readonly

    def __del__(self):
        self._command_stack.clear()

    def set(self, type, offset, value):
        """
        Set the specified bit field and returns its old value.
        """

        if self.readonly:
            raise ReadOnlyError()

        self._command_stack.extend(["SET", type, offset, value])

        return self

    def get(self, type, offset):
        """
        Returns the specified bit field.
        """

        self._command_stack.extend(["GET", type, offset])

        return self

    def incrby(self, type, offset, increment):
        """
        Increments or decrements (if a negative increment is given)
        the specified bit field and returns the new value.
        """

        if self.readonly:
            raise ReadOnlyError()

        self._command_stack.extend(["INCRBY", type, offset, increment])

        return self

    def overflow(self, type="SAT"):
        """
        fine-tune the behavior of the increment or decrement overflow,
        have no effect unless used before `incrby`
        three types are available: WRAP|SAT|FAIL
        """

        if self.readonly:
            raise ReadOnlyError()
        self._command_stack.extend(["OVERFLOW", type])

        return self

    async def exc(self):
        """execute commands in command stack"""

        return await self.redis.execute_command(*self._command_stack)


class StringsCommandMixin:
    RESPONSE_CALLBACKS = dict_merge(
        string_keys_to_dict("MSETNX PSETEX SETEX SETNX", bool),
        string_keys_to_dict(
            "BITCOUNT BITPOS DECRBY GETBIT INCRBY " "STRLEN SETBIT", int
        ),
        {
            "INCRBYFLOAT": float,
            "MSET": bool_ok,
            "SET": lambda r: r and nativestr(r) == "OK",
        },
    )

    async def append(self, key, value):
        """
        Appends the string ``value`` to the value at ``key``. If ``key``
        doesn't already exist, create it with a value of ``value``.
        Returns the new length of the value at ``key``.
        """

        return await self.execute_command("APPEND", key, value)

    async def bitcount(self, key, start=None, end=None):
        """
        Returns the count of set bits in the value of ``key``.  Optional
        ``start`` and ``end`` paramaters indicate which bytes to consider
        """
        params = [key]

        if start is not None and end is not None:
            params.append(start)
            params.append(end)
        elif (start is not None and end is None) or (end is not None and start is None):
            raise RedisError("Both start and end must be specified")

        return await self.execute_command("BITCOUNT", *params)

    async def bitop(self, operation, dest, *keys):
        """
        Perform a bitwise operation using ``operation`` between ``keys`` and
        store the result in ``dest``.
        """

        return await self.execute_command("BITOP", operation, dest, *keys)

    async def bitpos(self, key, bit, start=None, end=None):
        """
        Return the position of the first bit set to 1 or 0 in a string.
        ``start`` and ``end`` difines search range. The range is interpreted
        as a range of bytes and not a range of bits, so start=0 and end=2
        means to look at the first three bytes.
        """

        if bit not in (0, 1):
            raise RedisError("bit must be 0 or 1")
        params = [key, bit]

        start is not None and params.append(start)

        if start is not None and end is not None:
            params.append(end)
        elif start is None and end is not None:
            raise RedisError("start argument is not set, " "when end is specified")

        return await self.execute_command("BITPOS", *params)

    def bitfield(self, key):
        """
        Return a :class:`BitFieldOperation` instance to conveniently construct one or
        more bitfield operations on ``key``.
        """

        return BitFieldOperation(self, key)

    def bitfield_ro(self, key):
        """
        Return a :class:`BitFieldOperation` instance to conveniently construct bitfield
        operations on a read only replica against ``key``.

        Raises :class:`ReadOnlyError` if a write operation is attempted

        .. versionadded:: 2.1.0
        """

        return BitFieldOperation(self, key, readonly=True)

    async def decr(self, name, amount=1):
        """
        Decrements the value of ``key`` by ``amount``.  If no key exists,
        the value will be initialized as 0 - ``amount``
        """

        return await self.execute_command("DECRBY", name, amount)

    async def decrby(self, name, amount=1):
        """
        Decrements the value of ``key`` by ``amount``.  If no key exists,
        the value will be initialized as 0 - ``amount``

        .. versionadded:: 2.1.0
        """

        return await self.execute_command("DECRBY", name, amount)

    async def get(self, name):
        """
        Return the value at key ``name``, or None if the key doesn't exist
        """

        return await self.execute_command("GET", name)

    async def getdel(self, name):
        """
        Get the value at key ``name`` and delete the key. This command
        is similar to GET, except for the fact that it also deletes
        the key on success (if and only if the key's value type
        is a string).

        .. versionadded:: 2.1.0
        """

        return await self.execute_command("GETDEL", name)

    async def getex(self, name, ex=None, px=None, exat=None, pxat=None, persist=False):
        """
        Get the value of key and optionally set its expiration.

        GETEX is similar to GET, but is a write command with
        additional options. All time parameters can be given as
        :class:`datetime.timedelta` or integers.

        :param name: name of the key
        :param ex: sets an expire flag on key ``name`` for ``ex`` seconds.
        :param px: sets an expire flag on key ``name`` for ``px`` milliseconds.
        :param exat: sets an expire flag on key ``name`` for ``ex`` seconds,
         specified in unix time.
        :param pxat: sets an expire flag on key ``name`` for ``ex`` milliseconds,
         specified in unix time.
        :param persist: remove the time to live associated with ``name``.

        .. versionadded:: 2.1.0
        """

        opset = {ex, px, exat, pxat}

        if len(opset) > 2 or len(opset) > 1 and persist:
            raise DataError(
                "``ex``, ``px``, ``exat``, ``pxat``, "
                "and ``persist`` are mutually exclusive."
            )

        pieces = []
        # similar to set command

        if ex is not None:
            pieces.append("EX")

            if isinstance(ex, datetime.timedelta):
                ex = int(ex.total_seconds())
            pieces.append(ex)

        if px is not None:
            pieces.append("PX")

            if isinstance(px, datetime.timedelta):
                px = int(px.total_seconds() * 1000)
            pieces.append(px)
        # similar to pexpireat command

        if exat is not None:
            pieces.append("EXAT")

            if isinstance(exat, datetime.datetime):
                s = int(exat.microsecond / 1000000)
                exat = int(time.mktime(exat.timetuple())) + s
            pieces.append(exat)

        if pxat is not None:
            pieces.append("PXAT")

            if isinstance(pxat, datetime.datetime):
                ms = int(pxat.microsecond / 1000)
                pxat = int(time.mktime(pxat.timetuple())) * 1000 + ms
            pieces.append(pxat)

        if persist:
            pieces.append("PERSIST")

        return await self.execute_command("GETEX", name, *pieces)

    async def getbit(self, name, offset):
        "Returns a boolean indicating the value of ``offset`` in ``name``"

        return await self.execute_command("GETBIT", name, offset)

    async def getrange(self, key, start, end):
        """
        Returns the substring of the string value stored at ``key``,
        determined by the offsets ``start`` and ``end`` (both are inclusive)
        """

        return await self.execute_command("GETRANGE", key, start, end)

    async def getset(self, name, value):
        """
        Sets the value at key ``name`` to ``value``
        and returns the old value at key ``name`` atomically.
        """

        return await self.execute_command("GETSET", name, value)

    async def incr(self, name, amount=1):
        """
        Increments the value of ``key`` by ``amount``.  If no key exists,
        the value will be initialized as ``amount``
        """

        return await self.execute_command("INCRBY", name, amount)

    async def incrby(self, name, amount=1):
        """
        Increments the value of ``key`` by ``amount``.  If no key exists,
        the value will be initialized as ``amount``
        """

        # An alias for ``incr()``, because it is already implemented
        # as INCRBY redis command.

        return await self.incr(name, amount)

    async def incrbyfloat(self, name, amount=1.0):
        """
        Increments the value at key ``name`` by floating ``amount``.
        If no key exists, the value will be initialized as ``amount``
        """

        return await self.execute_command("INCRBYFLOAT", name, amount)

    async def mget(self, keys, *args):
        """
        Returns a list of values ordered identically to ``keys``
        """
        args = list_or_args(keys, args)

        return await self.execute_command("MGET", *args)

    async def mset(self, *args, **kwargs):
        """
        Sets key/values based on a mapping. Mapping can be supplied as a single
        dictionary argument or as kwargs.
        """

        if args:
            if len(args) != 1 or not isinstance(args[0], dict):
                raise RedisError("MSET requires **kwargs or a single dict arg")
            kwargs.update(args[0])
        items = []

        for pair in iteritems(kwargs):
            items.extend(pair)

        return await self.execute_command("MSET", *items)

    async def msetnx(self, *args, **kwargs):
        """
        Sets key/values based on a mapping if none of the keys are already set.
        Mapping can be supplied as a single dictionary argument or as kwargs.
        Returns a boolean indicating if the operation was successful.
        """

        if args:
            if len(args) != 1 or not isinstance(args[0], dict):
                raise RedisError("MSETNX requires **kwargs or a single " "dict arg")
            kwargs.update(args[0])
        items = []

        for pair in iteritems(kwargs):
            items.extend(pair)

        return await self.execute_command("MSETNX", *items)

    async def psetex(self, name, time_ms, value):
        """
        Set the value of key ``name`` to ``value`` that expires in ``time_ms``
        milliseconds. ``time_ms`` can be represented by an integer or a Python
        timedelta object
        """

        if isinstance(time_ms, datetime.timedelta):
            ms = int(time_ms.microseconds / 1000)
            time_ms = (time_ms.seconds + time_ms.days * 24 * 3600) * 1000 + ms

        return await self.execute_command("PSETEX", name, time_ms, value)

    async def set(self, name, value, ex=None, px=None, nx=False, xx=False):
        """
        Set the value at key ``name`` to ``value``

        ``ex`` sets an expire flag on key ``name`` for ``ex`` seconds.

        ``px`` sets an expire flag on key ``name`` for ``px`` milliseconds.

        ``nx`` if set to True, set the value at key ``name`` to ``value`` if it
            does not already exist.

        ``xx`` if set to True, set the value at key ``name`` to ``value`` if it
            already exists.
        """
        pieces = [name, value]

        if ex is not None:
            pieces.append("EX")

            if isinstance(ex, datetime.timedelta):
                ex = ex.seconds + ex.days * 24 * 3600
            pieces.append(ex)

        if px is not None:
            pieces.append("PX")

            if isinstance(px, datetime.timedelta):
                ms = int(px.microseconds / 1000)
                px = (px.seconds + px.days * 24 * 3600) * 1000 + ms
            pieces.append(px)

        if nx:
            pieces.append("NX")

        if xx:
            pieces.append("XX")

        return await self.execute_command("SET", *pieces)

    async def setbit(self, name, offset, value):
        """
        Flag the ``offset`` in ``name`` as ``value``. Returns a boolean
        indicating the previous value of ``offset``.
        """
        value = value and 1 or 0

        return await self.execute_command("SETBIT", name, offset, value)

    async def setex(self, name, time, value):
        """
        Set the value of key ``name`` to ``value`` that expires in ``time``
        seconds. ``time`` can be represented by an integer or a Python
        timedelta object.
        """

        if isinstance(time, datetime.timedelta):
            time = time.seconds + time.days * 24 * 3600

        return await self.execute_command("SETEX", name, time, value)

    async def setnx(self, name, value):
        """
        Sets the value of key ``name`` to ``value`` if key doesn't exist
        """

        return await self.execute_command("SETNX", name, value)

    async def setrange(self, name, offset, value):
        """
        Overwrite bytes in the value of ``name`` starting at ``offset`` with
        ``value``. If ``offset`` plus the length of ``value`` exceeds the
        length of the original value, the new value will be larger than before.
        If ``offset`` exceeds the length of the original value, null bytes
        will be used to pad between the end of the previous value and the start
        of what's being injected.

        Returns the length of the new string.
        """

        return await self.execute_command("SETRANGE", name, offset, value)

    async def strlen(self, name):
        """Returns the number of bytes stored in the value of ``name``"""

        return await self.execute_command("STRLEN", name)

    async def substr(self, name, start, end=-1):
        """
        Returns a substring of the string at key ``name``. ``start`` and ``end``
        are 0-based integers specifying the portion of the string to return.
        """

        return await self.execute_command("SUBSTR", name, start, end)


class ClusterStringsCommandMixin(StringsCommandMixin):

    NODES_FLAGS = {"BITOP": NodeFlag.BLOCKED}

    async def mget(self, keys, *args):
        """
        Returns a list of values ordered identically to ``keys``

        Cluster impl:
            Itterate all keys and send GET for each key.
            This will go alot slower than a normal mget call in StrictRedis.

            Operation is no longer atomic.
        """
        res = list()

        for arg in list_or_args(keys, args):
            res.append(await self.get(arg))

        return res

    async def mset(self, *args, **kwargs):
        """
        Sets key/values based on a mapping. Mapping can be supplied as a single
        dictionary argument or as kwargs.

        Cluster impl:
            Itterate over all items and do SET on each (k,v) pair

            Operation is no longer atomic.
        """

        if args:
            if len(args) != 1 or not isinstance(args[0], dict):
                raise RedisError("MSET requires **kwargs or a single dict arg")
            kwargs.update(args[0])

        for pair in iteritems(kwargs):
            await self.set(pair[0], pair[1])

        return True

    async def msetnx(self, *args, **kwargs):
        """
        Sets key/values based on a mapping if none of the keys are already set.
        Mapping can be supplied as a single dictionary argument or as kwargs.
        Returns a boolean indicating if the operation was successful.

        Clutser impl:
            Itterate over all items and do GET to determine if all keys do not exists.
            If true then call mset() on all keys.
        """

        if args:
            if len(args) != 1 or not isinstance(args[0], dict):
                raise RedisError("MSETNX requires **kwargs or a single dict arg")
            kwargs.update(args[0])

        # Itterate over all items and fail fast if one value is True.

        for k, _ in kwargs.items():
            if await self.get(k):
                return False

        return await self.mset(**kwargs)



import sys
from functools import wraps

from coredis.exceptions import ClusterDownError, RedisClusterException

_C_EXTENSION_SPEEDUP = False
try:
    from coredis.speedups import crc16, hash_slot

    _C_EXTENSION_SPEEDUP = True
except Exception:
    pass

LOOP_DEPRECATED = sys.version_info >= (3, 8)


def b(x):
    return x.encode("latin-1") if not isinstance(x, bytes) else x


def nativestr(x):
    return x if isinstance(x, str) else x.decode("utf-8", "replace")


def iteritems(x):
    return iter(x.items())


def iterkeys(x):
    return iter(x.keys())


def itervalues(x):
    return iter(x.values())


def ban_python_version_lt(min_version):
    min_version = tuple(map(int, min_version.split(".")))

    def decorator(func):
        @wraps(func)
        def _inner(*args, **kwargs):
            if sys.version_info[:2] < min_version:
                raise EnvironmentError(
                    "{} not supported in Python version less than {}".format(
                        func.__name__, min_version
                    )
                )
            else:
                return func(*args, **kwargs)

        return _inner

    return decorator


class dummy:
    """
    Instances of this class can be used as an attribute container.
    """

    def __init__(self):
        self.token = None

    def set(self, value):
        self.token = value

    def get(self):
        return self.token


# ++++++++++ response callbacks ++++++++++++++
def string_keys_to_dict(key_string, callback):
    return dict.fromkeys(key_string.split(), callback)


def list_keys_to_dict(key_list, callback):
    return dict.fromkeys(key_list, callback)


def dict_merge(*dicts):
    merged = {}
    for d in dicts:
        merged.update(d)
    return merged


def bool_ok(response):
    return nativestr(response) == "OK"


def list_or_args(keys, args):
    # returns a single list combining keys and args
    try:
        iter(keys)
        # a string or bytes instance can be iterated, but indicates
        # keys wasn't passed as a list
        if isinstance(keys, (str, bytes)):
            keys = [keys]
    except TypeError:
        keys = [keys]
    if args:
        keys.extend(args)
    return keys


def int_or_none(response):
    if response is None:
        return None
    return int(response)


def pairs_to_dict(response):
    """Creates a dict given a list of key/value pairs"""
    it = iter(response)
    return dict(zip(it, it))


# ++++++++++ result callbacks (cluster)++++++++++++++
def merge_result(res):
    """
    Merges all items in `res` into a list.

    This command is used when sending a command to multiple nodes
    and they result from each node should be merged into a single list.
    """
    if not isinstance(res, dict):
        raise ValueError("Value should be of dict type")

    result = set([])

    for _, v in res.items():
        for value in v:
            result.add(value)

    return list(result)


def first_key(res):
    """
    Returns the first result for the given command.

    If more then 1 result is returned then a `RedisClusterException` is raised.
    """
    if not isinstance(res, dict):
        raise ValueError("Value should be of dict type")

    if len(res.keys()) != 1:
        raise RedisClusterException("More then 1 result from command")

    return list(res.values())[0]


def blocked_command(self, command):
    """
    Raises a `RedisClusterException` mentioning the command is blocked.
    """
    raise RedisClusterException(
        "Command: {0} is blocked in redis cluster mode".format(command)
    )


def clusterdown_wrapper(func):
    """
    Wrapper for CLUSTERDOWN error handling.

    If the cluster reports it is down it is assumed that:
     - connection_pool was disconnected
     - connection_pool was reseted
     - refereh_table_asap set to True

    It will try 3 times to rerun the command and raises ClusterDownException if it continues to
    fail.
    """

    @wraps(func)
    async def inner(*args, **kwargs):
        for _ in range(0, 3):
            try:
                return await func(*args, **kwargs)
            except ClusterDownError:
                # Try again with the new cluster setup. All other errors
                # should be raised.
                pass

        # If it fails 3 times then raise exception back to caller
        raise ClusterDownError("CLUSTERDOWN error. Unable to rebuild the cluster")

    return inner


if not _C_EXTENSION_SPEEDUP:
    x_mode_m_crc16_lookup = [
        0x0000,
        0x1021,
        0x2042,
        0x3063,
        0x4084,
        0x50A5,
        0x60C6,
        0x70E7,
        0x8108,
        0x9129,
        0xA14A,
        0xB16B,
        0xC18C,
        0xD1AD,
        0xE1CE,
        0xF1EF,
        0x1231,
        0x0210,
        0x3273,
        0x2252,
        0x52B5,
        0x4294,
        0x72F7,
        0x62D6,
        0x9339,
        0x8318,
        0xB37B,
        0xA35A,
        0xD3BD,
        0xC39C,
        0xF3FF,
        0xE3DE,
        0x2462,
        0x3443,
        0x0420,
        0x1401,
        0x64E6,
        0x74C7,
        0x44A4,
        0x5485,
        0xA56A,
        0xB54B,
        0x8528,
        0x9509,
        0xE5EE,
        0xF5CF,
        0xC5AC,
        0xD58D,
        0x3653,
        0x2672,
        0x1611,
        0x0630,
        0x76D7,
        0x66F6,
        0x5695,
        0x46B4,
        0xB75B,
        0xA77A,
        0x9719,
        0x8738,
        0xF7DF,
        0xE7FE,
        0xD79D,
        0xC7BC,
        0x48C4,
        0x58E5,
        0x6886,
        0x78A7,
        0x0840,
        0x1861,
        0x2802,
        0x3823,
        0xC9CC,
        0xD9ED,
        0xE98E,
        0xF9AF,
        0x8948,
        0x9969,
        0xA90A,
        0xB92B,
        0x5AF5,
        0x4AD4,
        0x7AB7,
        0x6A96,
        0x1A71,
        0x0A50,
        0x3A33,
        0x2A12,
        0xDBFD,
        0xCBDC,
        0xFBBF,
        0xEB9E,
        0x9B79,
        0x8B58,
        0xBB3B,
        0xAB1A,
        0x6CA6,
        0x7C87,
        0x4CE4,
        0x5CC5,
        0x2C22,
        0x3C03,
        0x0C60,
        0x1C41,
        0xEDAE,
        0xFD8F,
        0xCDEC,
        0xDDCD,
        0xAD2A,
        0xBD0B,
        0x8D68,
        0x9D49,
        0x7E97,
        0x6EB6,
        0x5ED5,
        0x4EF4,
        0x3E13,
        0x2E32,
        0x1E51,
        0x0E70,
        0xFF9F,
        0xEFBE,
        0xDFDD,
        0xCFFC,
        0xBF1B,
        0xAF3A,
        0x9F59,
        0x8F78,
        0x9188,
        0x81A9,
        0xB1CA,
        0xA1EB,
        0xD10C,
        0xC12D,
        0xF14E,
        0xE16F,
        0x1080,
        0x00A1,
        0x30C2,
        0x20E3,
        0x5004,
        0x4025,
        0x7046,
        0x6067,
        0x83B9,
        0x9398,
        0xA3FB,
        0xB3DA,
        0xC33D,
        0xD31C,
        0xE37F,
        0xF35E,
        0x02B1,
        0x1290,
        0x22F3,
        0x32D2,
        0x4235,
        0x5214,
        0x6277,
        0x7256,
        0xB5EA,
        0xA5CB,
        0x95A8,
        0x8589,
        0xF56E,
        0xE54F,
        0xD52C,
        0xC50D,
        0x34E2,
        0x24C3,
        0x14A0,
        0x0481,
        0x7466,
        0x6447,
        0x5424,
        0x4405,
        0xA7DB,
        0xB7FA,
        0x8799,
        0x97B8,
        0xE75F,
        0xF77E,
        0xC71D,
        0xD73C,
        0x26D3,
        0x36F2,
        0x0691,
        0x16B0,
        0x6657,
        0x7676,
        0x4615,
        0x5634,
        0xD94C,
        0xC96D,
        0xF90E,
        0xE92F,
        0x99C8,
        0x89E9,
        0xB98A,
        0xA9AB,
        0x5844,
        0x4865,
        0x7806,
        0x6827,
        0x18C0,
        0x08E1,
        0x3882,
        0x28A3,
        0xCB7D,
        0xDB5C,
        0xEB3F,
        0xFB1E,
        0x8BF9,
        0x9BD8,
        0xABBB,
        0xBB9A,
        0x4A75,
        0x5A54,
        0x6A37,
        0x7A16,
        0x0AF1,
        0x1AD0,
        0x2AB3,
        0x3A92,
        0xFD2E,
        0xED0F,
        0xDD6C,
        0xCD4D,
        0xBDAA,
        0xAD8B,
        0x9DE8,
        0x8DC9,
        0x7C26,
        0x6C07,
        0x5C64,
        0x4C45,
        0x3CA2,
        0x2C83,
        0x1CE0,
        0x0CC1,
        0xEF1F,
        0xFF3E,
        0xCF5D,
        0xDF7C,
        0xAF9B,
        0xBFBA,
        0x8FD9,
        0x9FF8,
        0x6E17,
        0x7E36,
        0x4E55,
        0x5E74,
        0x2E93,
        0x3EB2,
        0x0ED1,
        0x1EF0,
    ]

    def _crc16(data):
        crc = 0
        for byte in data:
            crc = ((crc << 8) & 0xFF00) ^ x_mode_m_crc16_lookup[
                ((crc >> 8) & 0xFF) ^ byte
            ]
        return crc & 0xFFFF

    crc16 = _crc16  # noqa: F811

    def _hash_slot(key):
        start = key.find(b"{")
        if start > -1:
            end = key.find(b"}", start + 1)
            if end > -1 and end != start + 1:
                key = key[start + 1 : end]
        return crc16(key) % 16384

    hash_slot = _hash_slot  # noqa: F811


class NodeFlag:
    BLOCKED = "blocked"
    ALL_NODES = "all-nodes"
    ALL_MASTERS = "all-masters"
    RANDOM = "random"
    SLOT_ID = "slot-id"



import asyncio
import threading

from coredis.exceptions import ConnectionError, PubSubError, TimeoutError
from coredis.utils import iteritems, iterkeys, list_or_args, nativestr


class PubSub:
    """
    PubSub provides publish, subscribe and listen support to Redis channels.

    After subscribing to one or more channels, the listen() method will block
    until a message arrives on one of the subscribed channels. That message
    will be returned and it's safe to start listening again.
    """

    PUBLISH_MESSAGE_TYPES = ("message", "pmessage")
    UNSUBSCRIBE_MESSAGE_TYPES = ("unsubscribe", "punsubscribe")

    def __init__(self, connection_pool, ignore_subscribe_messages=False):
        self.connection_pool = connection_pool
        self.ignore_subscribe_messages = ignore_subscribe_messages
        self.connection = None
        # we need to know the encoding options for this connection in order
        # to lookup channel and pattern names for callback handlers.
        conn = connection_pool.get_connection("pubsub")
        try:
            self.encoding = conn.encoding
            self.decode_responses = conn.decode_responses
        finally:
            connection_pool.release(conn)
        self.reset()

    def __del__(self):
        try:
            # if this object went out of scope prior to shutting down
            # subscriptions, close the connection manually before
            # returning it to the connection pool
            self.reset()
        except Exception:
            pass

    def reset(self):
        if self.connection:
            self.connection.disconnect()
            self.connection.clear_connect_callbacks()
            self.connection_pool.release(self.connection)
            self.connection = None
        self.channels = {}
        self.patterns = {}

    def close(self):
        self.reset()

    async def on_connect(self, connection):
        """Re-subscribe to any channels and patterns previously subscribed to"""
        if self.channels:
            channels = {}

            for k, v in iteritems(self.channels):
                if not self.decode_responses:
                    k = k.decode(self.encoding)
                channels[k] = v
            await self.subscribe(**channels)

        if self.patterns:
            patterns = {}

            for k, v in iteritems(self.patterns):
                if not self.decode_responses:
                    k = k.decode(self.encoding)
                patterns[k] = v
            await self.psubscribe(**patterns)

    def encode(self, value):
        """
        Encodes the value so that it's identical to what we'll read off the
        connection
        """

        if self.decode_responses and isinstance(value, bytes):
            value = value.decode(self.encoding)
        elif not self.decode_responses and isinstance(value, str):
            value = value.encode(self.encoding)

        return value

    @property
    def subscribed(self):
        """Indicates if there are subscriptions to any channels or patterns"""

        return bool(self.channels or self.patterns)

    async def execute_command(self, *args, **kwargs):
        """Executes a publish/subscribe command"""

        # NOTE: don't parse the response in this function -- it could pull a
        # legitimate message off the stack if the connection is already
        # subscribed to one or more channels

        if self.connection is None:
            self.connection = self.connection_pool.get_connection()
            # register a callback that re-subscribes to any channels we
            # were listening to when we were disconnected
            self.connection.register_connect_callback(self.on_connect)
        connection = self.connection
        await self._execute(connection, connection.send_command, *args)

    async def _execute(self, connection, command, *args):
        try:
            return await command(*args)
        except asyncio.CancelledError:
            # do not retry if coroutine is cancelled

            if await connection.can_read():
                # disconnect if buffer is not empty in case of error
                # when connection is reused
                connection.disconnect()

            return None
        except (ConnectionError, TimeoutError) as e:
            connection.disconnect()

            if not connection.retry_on_timeout and isinstance(e, TimeoutError):
                raise
            # Connect manually here. If the Redis server is down, this will
            # fail and raise a ConnectionError as desired.
            await connection.connect()
            # the ``on_connect`` callback should haven been called by the
            # connection to resubscribe us to any channels and patterns we were
            # previously listening to

            return await command(*args)

    async def parse_response(self, block=True, timeout=0):
        """Parses the response from a publish/subscribe command"""
        connection = self.connection

        if connection is None:
            raise RuntimeError(
                "pubsub connection not set: "
                "did you forget to call subscribe() or psubscribe()?"
            )
        coro = self._execute(connection, connection.read_response)

        if not block and timeout > 0:
            try:
                return await asyncio.wait_for(coro, timeout)
            except Exception:
                return None

        return await coro

    async def psubscribe(self, *args, **kwargs):
        """
        Subscribes to channel patterns. Patterns supplied as keyword arguments
        expect a pattern name as the key and a callable as the value. A
        pattern's callable will be invoked automatically when a message is
        received on that pattern rather than producing a message via
        ``listen()``.
        """

        if args:
            args = list_or_args(args[0], args[1:])
        new_patterns = {}
        new_patterns.update(dict.fromkeys(map(self.encode, args)))

        for pattern, handler in iteritems(kwargs):
            new_patterns[self.encode(pattern)] = handler
        ret_val = await self.execute_command("PSUBSCRIBE", *iterkeys(new_patterns))
        # update the patterns dict AFTER we send the command. we don't want to
        # subscribe twice to these patterns, once for the command and again
        # for the reconnection.
        self.patterns.update(new_patterns)

        return ret_val

    async def punsubscribe(self, *args):
        """
        Unsubscribes from the supplied patterns. If empy, unsubscribe from
        all patterns.
        """

        if args:
            args = list_or_args(args[0], args[1:])

        return await self.execute_command("PUNSUBSCRIBE", *args)

    async def subscribe(self, *args, **kwargs):
        """
        Subscribes to channels. Channels supplied as keyword arguments expect
        a channel name as the key and a callable as the value. A channel's
        callable will be invoked automatically when a message is received on
        that channel rather than producing a message via ``listen()`` or
        ``get_message()``.
        """

        if args:
            args = list_or_args(args[0], args[1:])
        new_channels = {}
        new_channels.update(dict.fromkeys(map(self.encode, args)))

        for channel, handler in iteritems(kwargs):
            new_channels[self.encode(channel)] = handler
        ret_val = await self.execute_command("SUBSCRIBE", *iterkeys(new_channels))
        # update the channels dict AFTER we send the command. we don't want to
        # subscribe twice to these channels, once for the command and again
        # for the reconnection.
        self.channels.update(new_channels)

        return ret_val

    async def unsubscribe(self, *args):
        """
        Unsubscribes from the supplied channels. If empty, unsubscribe from
        all channels
        """

        if args:
            args = list_or_args(args[0], args[1:])

        return await self.execute_command("UNSUBSCRIBE", *args)

    async def listen(self):
        """
        Listens for messages on channels this client has been subscribed to
        """

        if self.subscribed:
            return self.handle_message(await self.parse_response(block=True))

    async def get_message(self, ignore_subscribe_messages=False, timeout=0):
        """
        Gets the next message if one is available, otherwise None.

        If timeout is specified, the system will wait for `timeout` seconds
        before returning. Timeout should be specified as a floating point
        number.
        """
        response = await self.parse_response(block=False, timeout=timeout)

        if response:
            return self.handle_message(response, ignore_subscribe_messages)

        return None

    def handle_message(self, response, ignore_subscribe_messages=False):
        """
        Parses a pub/sub message. If the channel or pattern was subscribed to
        with a message handler, the handler is invoked instead of a parsed
        message being returned.
        """
        message_type = nativestr(response[0])

        if message_type == "pmessage":
            message = {
                "type": message_type,
                "pattern": response[1],
                "channel": response[2],
                "data": response[3],
            }
        else:
            message = {
                "type": message_type,
                "pattern": None,
                "channel": response[1],
                "data": response[2],
            }

        # if this is an unsubscribe message, remove it from memory

        if message_type in self.UNSUBSCRIBE_MESSAGE_TYPES:
            subscribed_dict = None

            if message_type == "punsubscribe":
                subscribed_dict = self.patterns
            else:
                subscribed_dict = self.channels
            try:
                del subscribed_dict[message["channel"]]
            except KeyError:
                pass

        if message_type in self.PUBLISH_MESSAGE_TYPES:
            # if there's a message handler, invoke it
            handler = None

            if message_type == "pmessage":
                handler = self.patterns.get(message["pattern"], None)
            else:
                handler = self.channels.get(message["channel"], None)

            if handler:
                handler(message)

                return None
        else:
            # this is a subscribe/unsubscribe message. ignore if we don't
            # want them

            if ignore_subscribe_messages or self.ignore_subscribe_messages:
                return None

        return message

    def run_in_thread(self, daemon=False, poll_timeout=1.0):
        for channel, handler in iteritems(self.channels):
            if handler is None:
                raise PubSubError(
                    "Channel: '{}' has no handler registered".format(channel)
                )

        for pattern, handler in iteritems(self.patterns):
            if handler is None:
                raise PubSubError(
                    "Pattern: '{}' has no handler registered".format(pattern)
                )
        thread = PubSubWorkerThread(self, daemon=daemon, poll_timeout=poll_timeout)
        thread.start()

        return thread


class PubSubWorkerThread(threading.Thread):
    def __init__(self, pubsub, daemon=False, poll_timeout=1.0):
        super(PubSubWorkerThread, self).__init__()
        self.daemon = daemon
        self.pubsub = pubsub
        self.poll_timeout = poll_timeout
        self._running = False
        # Make sure we have the current thread loop before we
        # fork into the new thread. If not loop has been set on the connection
        # pool use the current default event loop.
        self.loop = pubsub.connection_pool.loop or asyncio.get_event_loop()

    async def _run(self):
        pubsub = self.pubsub

        while pubsub.subscribed:
            await pubsub.get_message(
                ignore_subscribe_messages=True, timeout=self.poll_timeout
            )
        pubsub.close()
        self._running = False

    def run(self):
        if self._running:
            return
        self._running = True
        future = asyncio.run_coroutine_threadsafe(self._run(), self.loop)

        return future.result()

    def stop(self):
        # stopping simply unsubscribes from all channels and patterns.
        # the unsubscribe responses that are generated will short circuit
        # the loop in run(), calling pubsub.close() to clean up the connection

        if self.loop:
            unsubscribed = asyncio.run_coroutine_threadsafe(
                self.pubsub.unsubscribe(), self.loop
            )
            punsubscribed = asyncio.run_coroutine_threadsafe(
                self.pubsub.punsubscribe(), self.loop
            )
            asyncio.wait([unsubscribed, punsubscribed], loop=self.loop)


class ClusterPubSub(PubSub):
    """Wrappers for the PubSub class"""

    def __init__(self, *args, **kwargs):
        super(ClusterPubSub, self).__init__(*args, **kwargs)

    async def execute_command(self, *args, **kwargs):
        """
        Executes a publish/subscribe command.

        NOTE: The code was initially taken from redis-py and tweaked to make it work within a
        cluster.
        """
        # NOTE: don't parse the response in this function -- it could pull a
        # legitimate message off the stack if the connection is already
        # subscribed to one or more channels
        await self.connection_pool.initialize()

        if self.connection is None:
            self.connection = self.connection_pool.get_connection(
                "pubsub",
                channel=args[1],
            )
            # register a callback that re-subscribes to any channels we
            # were listening to when we were disconnected
            self.connection.register_connect_callback(self.on_connect)
        connection = self.connection
        await self._execute(connection, connection.send_command, *args)



from coredis.cache import Cache, Compressor, IdentityGenerator, Serializer
from coredis.exceptions import ResponseError
from coredis.lock import Lock, LuaLock


class ExtraCommandMixin:

    RESPONSE_CALLBACKS = {}

    def cache(
        self,
        name,
        cache_class=Cache,
        identity_generator_class=IdentityGenerator,
        compressor_class=Compressor,
        serializer_class=Serializer,
        *args,
        **kwargs
    ):
        """
        Return a cache object using default identity generator,
        serializer and compressor.

        :param name: is used to identify the series of your cache
        :param cache_class: Cache is for normal use and HerdCache
         is used in case of Thundering Herd Problem
        :param identity_generator_class: is the class used to generate
         the real unique key in cache, can be overwritten to
         meet your special needs. It should provide `generate` API
        :param compressor_class: is the class used to compress cache in redis,
         can be overwritten with API `compress` and `decompress` retained.
        :param serializer_class: is the class used to serialize
         content before compress, can be overwritten with API
         `serialize` and `deserialize` retained.
        """
        return cache_class(
            self,
            app=name,
            identity_generator_class=identity_generator_class,
            compressor_class=compressor_class,
            serializer_class=serializer_class,
            *args,
            **kwargs
        )

    def lock(
        self,
        name,
        timeout=None,
        sleep=0.1,
        blocking_timeout=None,
        lock_class=None,
        thread_local=True,
    ):
        """
        Return a new Lock object using :paramref:`name` that mimics
        the behavior of threading.Lock.

        :param timeout: If specified indicates a maximum life for the lock.
         By default, it will remain locked until release() is called.
        :param sleep: indicates the amount of time to sleep per loop iteration
         when the lock is in blocking mode and another client is currently
         holding the lock.
        :param blocking_timeout: indicates the maximum amount of time in seconds to
         spend trying to acquire the lock. A value of ``None`` indicates
         continue trying forever. ``blocking_timeout`` can be specified as a
         float or integer, both representing the number of seconds to wait.
        :param lock_class: forces the specified lock implementation.
        :param thread_local: indicates whether the lock token is placed in
         thread-local storage.

         By default, the token is placed in thread local
         storage so that a thread only sees its token, not a token set by
         another thread. Consider the following timeline:

            time: 0, thread-1 acquires `my-lock`, with a timeout of 5 seconds.
                     thread-1 sets the token to "abc"
            time: 1, thread-2 blocks trying to acquire `my-lock` using the
                     Lock instance.
            time: 5, thread-1 has not yet completed. redis expires the lock
                     key.
            time: 5, thread-2 acquired `my-lock` now that it's available.
                     thread-2 sets the token to "xyz"
            time: 6, thread-1 finishes its work and calls release(). if the
                     token is *not* stored in thread local storage, then
                     thread-1 would see the token value as "xyz" and would be
                     able to successfully release the thread-2's lock.

        In some use cases it's necessary to disable thread local storage. For
        example, if you have code where one thread acquires a lock and passes
        that lock instance to a worker thread to release later. If thread
        local storage isn't disabled in this case, the worker thread won't see
        the token set by the thread that acquired the lock. Our assumption
        is that these cases aren't common and as such default to using
        thread local storage.
        """

        if lock_class is None:
            if self._use_lua_lock is None:
                # the first time .lock() is called, determine if we can use
                # Lua by attempting to register the necessary scripts
                try:
                    LuaLock.register_scripts(self)
                    self._use_lua_lock = True
                except ResponseError:
                    self._use_lua_lock = False
            lock_class = self._use_lua_lock and LuaLock or Lock
        return lock_class(
            self,
            name,
            timeout=timeout,
            sleep=sleep,
            blocking_timeout=blocking_timeout,
            thread_local=thread_local,
        )



from __future__ import with_statement

import asyncio
import os
import re

import pytest

import coredis
from coredis.exceptions import (
    BusyLoadingError,
    ConnectionError,
    ReadOnlyError,
    RedisError,
)
from coredis.pool import to_bool


class DummyConnection:
    description = "DummyConnection<>"

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.pid = os.getpid()
        self.awaiting_response = False


class TestConnectionPool:
    def get_pool(
        self,
        connection_kwargs=None,
        max_connections=None,
        connection_class=DummyConnection,
    ):
        connection_kwargs = connection_kwargs or {}
        pool = coredis.ConnectionPool(
            connection_class=connection_class,
            max_connections=max_connections,
            **connection_kwargs
        )
        return pool

    def test_connection_creation(self):
        connection_kwargs = {"foo": "bar", "biz": "baz"}
        pool = self.get_pool(connection_kwargs=connection_kwargs)
        connection = pool.get_connection()
        assert isinstance(connection, DummyConnection)
        assert connection.kwargs == connection_kwargs

    def test_multiple_connections(self):
        pool = self.get_pool()
        c1 = pool.get_connection()
        c2 = pool.get_connection()
        assert c1 != c2

    def test_max_connections(self):
        pool = self.get_pool(max_connections=2)
        pool.get_connection()
        pool.get_connection()
        with pytest.raises(ConnectionError):
            pool.get_connection()

    def test_reuse_previously_released_connection(self):
        pool = self.get_pool()
        c1 = pool.get_connection()
        pool.release(c1)
        c2 = pool.get_connection()
        assert c1 == c2

    def test_repr_contains_db_info_tcp(self):
        connection_kwargs = {"host": "localhost", "port": 6379, "db": 1}
        pool = self.get_pool(
            connection_kwargs=connection_kwargs, connection_class=coredis.Connection
        )
        expected = "ConnectionPool<Connection<host=localhost,port=6379,db=1>>"
        assert repr(pool) == expected

    def test_repr_contains_db_info_unix(self):
        connection_kwargs = {"path": "/abc", "db": 1}
        pool = self.get_pool(
            connection_kwargs=connection_kwargs,
            connection_class=coredis.UnixDomainSocketConnection,
        )
        expected = "ConnectionPool<UnixDomainSocketConnection<path=/abc,db=1>>"
        assert repr(pool) == expected

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_connection_idle_check(self, event_loop):
        rs = coredis.StrictRedis(
            host="127.0.0.1",
            port=6379,
            db=0,
            max_idle_time=0.2,
            idle_check_interval=0.1,
        )
        await rs.info()
        assert len(rs.connection_pool._available_connections) == 1
        assert len(rs.connection_pool._in_use_connections) == 0
        conn = rs.connection_pool._available_connections[0]
        last_active_at = conn.last_active_at
        await asyncio.sleep(0.3)
        assert len(rs.connection_pool._available_connections) == 0
        assert len(rs.connection_pool._in_use_connections) == 0
        assert last_active_at == conn.last_active_at
        assert conn._writer is None and conn._reader is None


class TestConnectionPoolURLParsing:
    def test_defaults(self):
        pool = coredis.ConnectionPool.from_url("redis://localhost")
        assert pool.connection_class == coredis.Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "port": 6379,
            "db": 0,
            "password": None,
        }

    def test_hostname(self):
        pool = coredis.ConnectionPool.from_url("redis://myhost")
        assert pool.connection_class == coredis.Connection
        assert pool.connection_kwargs == {
            "host": "myhost",
            "port": 6379,
            "db": 0,
            "password": None,
        }

    def test_quoted_hostname(self):
        pool = coredis.ConnectionPool.from_url(
            "redis://my %2F host %2B%3D+", decode_components=True
        )
        assert pool.connection_class == coredis.Connection
        assert pool.connection_kwargs == {
            "host": "my / host +=+",
            "port": 6379,
            "db": 0,
            "password": None,
        }

    def test_port(self):
        pool = coredis.ConnectionPool.from_url("redis://localhost:6380")
        assert pool.connection_class == coredis.Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "port": 6380,
            "db": 0,
            "password": None,
        }

    def test_password(self):
        pool = coredis.ConnectionPool.from_url("redis://:mypassword@localhost")
        assert pool.connection_class == coredis.Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "port": 6379,
            "db": 0,
            "password": "mypassword",
        }

    def test_quoted_password(self):
        pool = coredis.ConnectionPool.from_url(
            "redis://:%2Fmypass%2F%2B word%3D%24+@localhost", decode_components=True
        )
        assert pool.connection_class == coredis.Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "port": 6379,
            "db": 0,
            "password": "/mypass/+ word=$+",
        }

    def test_db_as_argument(self):
        pool = coredis.ConnectionPool.from_url("redis://localhost", db="1")
        assert pool.connection_class == coredis.Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "port": 6379,
            "db": 1,
            "password": None,
        }

    def test_db_in_path(self):
        pool = coredis.ConnectionPool.from_url("redis://localhost/2", db="1")
        assert pool.connection_class == coredis.Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "port": 6379,
            "db": 2,
            "password": None,
        }

    def test_db_in_querystring(self):
        pool = coredis.ConnectionPool.from_url("redis://localhost/2?db=3", db="1")
        assert pool.connection_class == coredis.Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "port": 6379,
            "db": 3,
            "password": None,
        }

    def test_extra_typed_querystring_options(self):
        pool = coredis.ConnectionPool.from_url(
            "redis://localhost/2?stream_timeout=20&connect_timeout=10"
        )

        assert pool.connection_class == coredis.Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "port": 6379,
            "db": 2,
            "stream_timeout": 20.0,
            "connect_timeout": 10.0,
            "password": None,
        }

    def test_boolean_parsing(self):
        for expected, value in (
            (None, None),
            (None, ""),
            (False, 0),
            (False, "0"),
            (False, "f"),
            (False, "F"),
            (False, "False"),
            (False, "n"),
            (False, "N"),
            (False, "No"),
            (True, 1),
            (True, "1"),
            (True, "y"),
            (True, "Y"),
            (True, "Yes"),
        ):
            assert expected is to_bool(value)

    def test_invalid_extra_typed_querystring_options(self):
        import warnings

        with warnings.catch_warnings(record=True) as warning_log:
            coredis.ConnectionPool.from_url(
                "redis://localhost/2?stream_timeout=_&" "connect_timeout=abc"
            )
        # Compare the message values
        assert [
            str(m.message) for m in sorted(warning_log, key=lambda l: str(l.message))
        ] == [
            "Invalid value for `connect_timeout` in connection URL.",
            "Invalid value for `stream_timeout` in connection URL.",
        ]

    def test_extra_querystring_options(self):
        pool = coredis.ConnectionPool.from_url("redis://localhost?a=1&b=2")
        assert pool.connection_class == coredis.Connection
        assert pool.connection_kwargs == {
            "host": "localhost",
            "port": 6379,
            "db": 0,
            "password": None,
            "a": "1",
            "b": "2",
        }

    def test_client_creates_connection_pool(self):
        r = coredis.StrictRedis.from_url("redis://myhost")
        assert r.connection_pool.connection_class == coredis.Connection
        assert r.connection_pool.connection_kwargs == {
            "host": "myhost",
            "port": 6379,
            "db": 0,
            "password": None,
        }


class TestConnectionPoolUnixSocketURLParsing:
    def test_defaults(self):
        pool = coredis.ConnectionPool.from_url("unix:///socket")
        assert pool.connection_class == coredis.UnixDomainSocketConnection
        assert pool.connection_kwargs == {
            "path": "/socket",
            "db": 0,
            "password": None,
        }

    def test_password(self):
        pool = coredis.ConnectionPool.from_url("unix://:mypassword@/socket")
        assert pool.connection_class == coredis.UnixDomainSocketConnection
        assert pool.connection_kwargs == {
            "path": "/socket",
            "db": 0,
            "password": "mypassword",
        }

    def test_quoted_password(self):
        pool = coredis.ConnectionPool.from_url(
            "unix://:%2Fmypass%2F%2B word%3D%24+@/socket", decode_components=True
        )
        assert pool.connection_class == coredis.UnixDomainSocketConnection
        assert pool.connection_kwargs == {
            "path": "/socket",
            "db": 0,
            "password": "/mypass/+ word=$+",
        }

    def test_quoted_path(self):
        pool = coredis.ConnectionPool.from_url(
            "unix://:mypassword@/my%2Fpath%2Fto%2F..%2F+_%2B%3D%24ocket",
            decode_components=True,
        )
        assert pool.connection_class == coredis.UnixDomainSocketConnection
        assert pool.connection_kwargs == {
            "path": "/my/path/to/../+_+=$ocket",
            "db": 0,
            "password": "mypassword",
        }

    def test_db_as_argument(self):
        pool = coredis.ConnectionPool.from_url("unix:///socket", db=1)
        assert pool.connection_class == coredis.UnixDomainSocketConnection
        assert pool.connection_kwargs == {
            "path": "/socket",
            "db": 1,
            "password": None,
        }

    def test_db_in_querystring(self):
        pool = coredis.ConnectionPool.from_url("unix:///socket?db=2", db=1)
        assert pool.connection_class == coredis.UnixDomainSocketConnection
        assert pool.connection_kwargs == {
            "path": "/socket",
            "db": 2,
            "password": None,
        }

    def test_extra_querystring_options(self):
        pool = coredis.ConnectionPool.from_url("unix:///socket?a=1&b=2")
        assert pool.connection_class == coredis.UnixDomainSocketConnection
        assert pool.connection_kwargs == {
            "path": "/socket",
            "db": 0,
            "password": None,
            "a": "1",
            "b": "2",
        }


class TestSSLConnectionURLParsing:
    def test_defaults(self):
        pool = coredis.ConnectionPool.from_url("rediss://localhost")
        assert pool.connection_class == coredis.Connection
        assert pool.connection_kwargs.pop("ssl_context") is not None
        assert pool.connection_kwargs == {
            "host": "localhost",
            "port": 6379,
            "db": 0,
            "password": None,
        }

    def test_cert_reqs_options(self):
        import ssl

        with pytest.raises(TypeError) as e:
            pool = coredis.ConnectionPool.from_url(
                "rediss://?ssl_cert_reqs=none&ssl_keyfile=test"
            )
            assert e.message == "certfile should be a valid filesystem path"
            assert pool.get_connection().ssl_context.verify_mode == ssl.CERT_NONE

        with pytest.raises(TypeError) as e:
            pool = coredis.ConnectionPool.from_url(
                "rediss://?ssl_cert_reqs=optional&ssl_keyfile=test"
            )
            assert e.message == "certfile should be a valid filesystem path"
            assert pool.get_connection().ssl_context.verify_mode == ssl.CERT_OPTIONAL

        with pytest.raises(TypeError) as e:
            pool = coredis.ConnectionPool.from_url(
                "rediss://?ssl_cert_reqs=required&ssl_keyfile=test"
            )
            assert e.message == "certfile should be a valid filesystem path"
            assert pool.get_connection().ssl_context.verify_mode == ssl.CERT_REQUIRED


class TestConnection:
    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_on_connect_error(self, event_loop):
        """
        An error in Connection.on_connect should disconnect from the server
        see for details: https://github.com/andymccurdy/redis-py/issues/368
        """
        # this assumes the Redis server being tested against doesn't have
        # 9999 databases ;)
        bad_connection = coredis.StrictRedis(db=9999, loop=event_loop)
        # an error should be raised on connect
        with pytest.raises(RedisError):
            await bad_connection.info()
        pool = bad_connection.connection_pool
        assert len(pool._available_connections) == 0

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_busy_loading_disconnects_socket(self, event_loop):
        """
        If Redis raises a LOADING error, the connection should be
        disconnected and a BusyLoadingError raised
        """
        client = coredis.StrictRedis(loop=event_loop)
        with pytest.raises(BusyLoadingError):
            await client.execute_command("DEBUG", "ERROR", "LOADING fake message")
        pool = client.connection_pool
        assert len(pool._available_connections) == 0

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_busy_loading_from_pipeline_immediate_command(self, event_loop):
        """
        BusyLoadingErrors should raise from Pipelines that execute a
        command immediately, like WATCH does.
        """
        client = coredis.StrictRedis(loop=event_loop)
        pipe = await client.pipeline()
        with pytest.raises(BusyLoadingError):
            await pipe.immediate_execute_command(
                "DEBUG", "ERROR", "LOADING fake message"
            )
        pool = client.connection_pool
        assert not pipe.connection
        assert len(pool._available_connections) == 0

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_busy_loading_from_pipeline(self, event_loop):
        """
        BusyLoadingErrors should be raised from a pipeline execution
        regardless of the raise_on_error flag.
        """
        client = coredis.StrictRedis(loop=event_loop)
        pipe = await client.pipeline()
        await pipe.execute_command("DEBUG", "ERROR", "LOADING fake message")
        with pytest.raises(RedisError):
            await pipe.execute()
        pool = client.connection_pool
        assert not pipe.connection
        assert len(pool._available_connections) == 1
        assert pool._available_connections[0]._writer
        assert pool._available_connections[0]._reader

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_read_only_error(self, event_loop):
        "READONLY errors get turned in ReadOnlyError exceptions"
        client = coredis.StrictRedis(loop=event_loop)
        with pytest.raises(ReadOnlyError):
            await client.execute_command("DEBUG", "ERROR", "READONLY blah blah")

    def test_connect_from_url_tcp(self):
        connection = coredis.StrictRedis.from_url("redis://localhost")
        pool = connection.connection_pool

        assert re.match("(.*)<(.*)<(.*)>>", repr(pool)).groups() == (
            "ConnectionPool",
            "Connection",
            "host=localhost,port=6379,db=0",
        )

    def test_connect_from_url_unix(self):
        connection = coredis.StrictRedis.from_url("unix:///path/to/socket")
        pool = connection.connection_pool

        assert re.match("(.*)<(.*)<(.*)>>", repr(pool)).groups() == (
            "ConnectionPool",
            "UnixDomainSocketConnection",
            "path=/path/to/socket,db=0",
        )



from coredis.utils import b, dict_merge, first_key, list_or_args, string_keys_to_dict


def parse_sscan(response, **options):
    cursor, r = response

    return int(cursor), r


class SetsCommandMixin:

    RESPONSE_CALLBACKS = dict_merge(
        string_keys_to_dict(
            "SADD SCARD SDIFFSTORE " "SETRANGE SINTERSTORE " "SREM SUNIONSTORE", int
        ),
        string_keys_to_dict("SISMEMBER SMOVE", bool),
        string_keys_to_dict(
            "SDIFF SINTER SMEMBERS SUNION", lambda r: r and set(r) or set()
        ),
        {"SSCAN": parse_sscan},
    )

    async def sadd(self, name, *values):
        """Adds ``value(s)`` to set ``name``"""

        return await self.execute_command("SADD", name, *values)

    async def scard(self, name):
        """Returns the number of elements in set ``name``"""

        return await self.execute_command("SCARD", name)

    async def sdiff(self, keys, *args):
        """Returns the difference of sets specified by ``keys``"""
        args = list_or_args(keys, args)

        return await self.execute_command("SDIFF", *args)

    async def sdiffstore(self, dest, keys, *args):
        """
        Stores the difference of sets specified by ``keys`` into a new
        set named ``dest``.  Returns the number of keys in the new set.
        """
        args = list_or_args(keys, args)

        return await self.execute_command("SDIFFSTORE", dest, *args)

    async def sinter(self, keys, *args):
        """Returns the intersection of sets specified by ``keys``"""
        args = list_or_args(keys, args)

        return await self.execute_command("SINTER", *args)

    async def sinterstore(self, dest, keys, *args):
        """
        Stores the intersection of sets specified by ``keys`` into a new
        set named ``dest``.  Returns the number of keys in the new set.
        """
        args = list_or_args(keys, args)

        return await self.execute_command("SINTERSTORE", dest, *args)

    async def sismember(self, name, value):
        """
        Returns a boolean indicating if ``value`` is a member of set ``name``
        """

        return await self.execute_command("SISMEMBER", name, value)

    async def smembers(self, name):
        """Returns all members of the set ``name``"""

        return await self.execute_command("SMEMBERS", name)

    async def smismember(self, name, values, *args):
        """
        Return whether each value in ``values`` is a member of the set ``name``
        as a list of ``bool`` in the order of ``values``

        .. versionadded:: 2.1.0
        """
        args = list_or_args(values, args)

        return await self.execute_command("SMISMEMBER", name, *args)

    async def smove(self, src, dst, value):
        """Moves ``value`` from set ``src`` to set ``dst`` atomically"""

        return await self.execute_command("SMOVE", src, dst, value)

    async def spop(self, name, count=None):
        """
        Removes and returns a random member of set ``name``
        ``count`` should be type of int and default set to 1.
        If ``count`` is supplied, pops a list of ``count`` random
        members of set ``name``
        """

        if count and isinstance(count, int):
            return await self.execute_command("SPOP", name, count)
        else:
            return await self.execute_command("SPOP", name)

    async def srandmember(self, name, number=None):
        """
        If ``number`` is None, returns a random member of set ``name``.

        If ``number`` is supplied, returns a list of ``number`` random
        memebers of set ``name``. Note this is only available when running
        Redis 2.6+.
        """
        args = number and [number] or []

        return await self.execute_command("SRANDMEMBER", name, *args)

    async def srem(self, name, *values):
        """Removes ``values`` from set ``name``"""

        return await self.execute_command("SREM", name, *values)

    async def sunion(self, keys, *args):
        """Returns the union of sets specified by ``keys``"""
        args = list_or_args(keys, args)

        return await self.execute_command("SUNION", *args)

    async def sunionstore(self, dest, keys, *args):
        """
        Stores the union of sets specified by ``keys`` into a new
        set named ``dest``.  Returns the number of keys in the new set.
        """
        args = list_or_args(keys, args)

        return await self.execute_command("SUNIONSTORE", dest, *args)

    async def sscan(self, name, cursor=0, match=None, count=None):
        """
        Incrementally returns lists of elements in a set. Also returns a
        cursor pointing to the scan position.

        ``match`` is for filtering the keys by pattern

        ``count`` is for hint the minimum number of returns
        """
        pieces = [name, cursor]

        if match is not None:
            pieces.extend([b("MATCH"), match])

        if count is not None:
            pieces.extend([b("COUNT"), count])

        return await self.execute_command("SSCAN", *pieces)


class ClusterSetsCommandMixin(SetsCommandMixin):

    RESULT_CALLBACKS = {"SSCAN": first_key}

    ###
    # Set commands

    async def sdiff(self, keys, *args):
        """
        Returns the difference of sets specified by ``keys``

        Cluster impl:
            Querry all keys and diff all sets and return result
        """
        k = list_or_args(keys, args)
        res = await self.smembers(k[0])

        for arg in k[1:]:
            res -= await self.smembers(arg)

        return res

    async def sdiffstore(self, dest, keys, *args):
        """
        Stores the difference of sets specified by ``keys`` into a new
        set named ``dest``.  Returns the number of keys in the new set.
        Overwrites dest key if it exists.

        Cluster impl:
            Use sdiff() --> Delete dest key --> store result in dest key
        """
        res = await self.sdiff(keys, *args)
        await self.delete(dest)

        if not res:
            return 0

        return await self.sadd(dest, *res)

    async def sinter(self, keys, *args):
        """
        Returns the intersection of sets specified by ``keys``

        Cluster impl:
            Querry all keys, intersection and return result
        """
        k = list_or_args(keys, args)
        res = await self.smembers(k[0])

        for arg in k[1:]:
            res &= await self.smembers(arg)

        return res

    async def sinterstore(self, dest, keys, *args):
        """
        Stores the intersection of sets specified by ``keys`` into a new
        set named ``dest``.  Returns the number of keys in the new set.

        Cluster impl:
            Use sinter() --> Delete dest key --> store result in dest key
        """
        res = await self.sinter(keys, *args)
        await self.delete(dest)

        if res:
            await self.sadd(dest, *res)

            return len(res)
        else:
            return 0

    async def smove(self, src, dst, value):
        """
        Moves ``value`` from set ``src`` to set ``dst`` atomically

        Cluster impl:
            SMEMBERS --> SREM --> SADD. Function is no longer atomic.
        """
        res = await self.srem(src, value)

        # Only add the element if existed in src set

        if res == 1:
            await self.sadd(dst, value)

        return res

    async def sunion(self, keys, *args):
        """
        Returns the union of sets specified by ``keys``

        Cluster impl:
            Querry all keys, union and return result

            Operation is no longer atomic.
        """
        k = list_or_args(keys, args)
        res = await self.smembers(k[0])

        for arg in k[1:]:
            res |= await self.smembers(arg)

        return res

    async def sunionstore(self, dest, keys, *args):
        """
        Stores the union of sets specified by ``keys`` into a new
        set named ``dest``.  Returns the number of keys in the new set.

        Cluster impl:
            Use sunion() --> Dlete dest key --> store result in dest key

            Operation is no longer atomic.
        """
        res = await self.sunion(keys, *args)
        await self.delete(dest)

        return await self.sadd(dest, *res)



import asyncio
import contextvars
import time as mod_time
import uuid
import warnings

from coredis.connection import ClusterConnection
from coredis.exceptions import LockError, WatchError
from coredis.utils import LOOP_DEPRECATED, b, dummy


class Lock:
    """
    A shared, distributed Lock. Using Redis for locking allows the Lock
    to be shared across processes and/or machines.

    It's left to the user to resolve deadlock issues and make sure
    multiple clients play nicely together.
    """

    def __init__(
        self,
        redis,
        name,
        timeout=None,
        sleep=0.1,
        blocking=True,
        blocking_timeout=None,
        thread_local=True,
    ):
        """
        Create a new Lock instance named ``name`` using the Redis client
        supplied by ``redis``.

        ``timeout`` indicates a maximum life for the lock.
        By default, it will remain locked until release() is called.
        ``timeout`` can be specified as a float or integer, both representing
        the number of seconds to wait.

        ``sleep`` indicates the amount of time to sleep per loop iteration
        when the lock is in blocking mode and another client is currently
        holding the lock.

        ``blocking`` indicates whether calling ``acquire`` should block until
        the lock has been acquired or to fail immediately, causing ``acquire``
        to return False and the lock not being acquired. Defaults to True.
        Note this value can be overridden by passing a ``blocking``
        argument to ``acquire``.

        ``blocking_timeout`` indicates the maximum amount of time in seconds to
        spend trying to acquire the lock. A value of ``None`` indicates
        continue trying forever. ``blocking_timeout`` can be specified as a
        float or integer, both representing the number of seconds to wait.

        ``thread_local`` indicates whether the lock token is placed in
        thread-local storage. By default, the token is placed in thread local
        storage so that a thread only sees its token, not a token set by
        another thread. Consider the following timeline:

            time: 0, thread-1 acquires `my-lock`, with a timeout of 5 seconds.
                     thread-1 sets the token to "abc"
            time: 1, thread-2 blocks trying to acquire `my-lock` using the
                     Lock instance.
            time: 5, thread-1 has not yet completed. redis expires the lock
                     key.
            time: 5, thread-2 acquired `my-lock` now that it's available.
                     thread-2 sets the token to "xyz"
            time: 6, thread-1 finishes its work and calls release(). if the
                     token is *not* stored in thread local storage, then
                     thread-1 would see the token value as "xyz" and would be
                     able to successfully release the thread-2's lock.

        In some use cases it's necessary to disable thread local storage. For
        example, if you have code where one thread acquires a lock and passes
        that lock instance to a worker thread to release later. If thread
        local storage isn't disabled in this case, the worker thread won't see
        the token set by the thread that acquired the lock. Our assumption
        is that these cases aren't common and as such default to using
        thread local storage.
        """
        self.redis = redis
        self.name = name
        self.timeout = timeout
        self.sleep = sleep
        self.blocking = blocking
        self.blocking_timeout = blocking_timeout
        self.thread_local = bool(thread_local)
        self.local = (
            contextvars.ContextVar("token", default=None)
            if self.thread_local
            else dummy()
        )
        if self.timeout and self.sleep > self.timeout:
            raise LockError("'sleep' must be less than 'timeout'")

    async def __aenter__(self):
        # force blocking, as otherwise the user would have to check whether
        # the lock was actually acquired or not.
        await self.acquire(blocking=True)
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.release()

    async def acquire(self, blocking=None, blocking_timeout=None):
        """
        Use Redis to hold a shared, distributed lock named ``name``.
        Returns True once the lock is acquired.

        If ``blocking`` is False, always return immediately. If the lock
        was acquired, return True, otherwise return False.

        ``blocking_timeout`` specifies the maximum number of seconds to
        wait trying to acquire the lock.
        """
        sleep = self.sleep
        token = b(uuid.uuid1().hex)
        if blocking is None:
            blocking = self.blocking
        if blocking_timeout is None:
            blocking_timeout = self.blocking_timeout
        stop_trying_at = None
        if blocking_timeout is not None:
            stop_trying_at = mod_time.time() + blocking_timeout
        while True:
            if await self.do_acquire(token):
                self.local.set(token)
                return True
            if not blocking:
                return False
            if stop_trying_at is not None and mod_time.time() > stop_trying_at:
                return False
            if LOOP_DEPRECATED:
                await asyncio.sleep(sleep)
            else:
                await asyncio.sleep(sleep, loop=self.redis.connection_pool.loop)

    async def do_acquire(self, token):
        if self.timeout:
            # convert to milliseconds
            timeout = int(self.timeout * 1000)
        else:
            timeout = None
        return await self.redis.set(self.name, token, nx=True, px=timeout)

    async def release(self):
        """Releases the already acquired lock"""
        expected_token = self.local.get()
        if expected_token is None:
            raise LockError("Cannot release an unlocked lock")
        self.local.set(None)
        await self.do_release(expected_token)

    async def do_release(self, expected_token):
        name = self.name

        async def execute_release(pipe):
            lock_value = await pipe.get(name)
            if lock_value != expected_token:
                raise LockError("Cannot release a lock that's no longer owned")
            await pipe.delete(name)

        await self.redis.transaction(execute_release, name)

    async def extend(self, additional_time):
        """
        Adds more time to an already acquired lock.

        ``additional_time`` can be specified as an integer or a float, both
        representing the number of seconds to add.
        """
        if self.local.get() is None:
            raise LockError("Cannot extend an unlocked lock")
        if self.timeout is None:
            raise LockError("Cannot extend a lock with no timeout")
        return await self.do_extend(additional_time)

    async def do_extend(self, additional_time):
        pipe = await self.redis.pipeline()
        await pipe.watch(self.name)
        lock_value = await pipe.get(self.name)
        if lock_value != self.local.get():
            raise LockError("Cannot extend a lock that's no longer owned")
        expiration = await pipe.pttl(self.name)
        if expiration is None or expiration < 0:
            # Redis evicted the lock key between the previous get() and now
            # we'll handle this when we call pexpire()
            expiration = 0
        pipe.multi()
        await pipe.pexpire(self.name, expiration + int(additional_time * 1000))

        try:
            response = await pipe.execute()
        except WatchError:
            # someone else acquired the lock
            raise LockError("Cannot extend a lock that's no longer owned")
        if not response[0]:
            # pexpire returns False if the key doesn't exist
            raise LockError("Cannot extend a lock that's no longer owned")
        return True


class LuaLock(Lock):
    """
    A lock implementation that uses Lua scripts rather than pipelines
    and watches.
    """

    lua_release = None
    lua_extend = None

    # KEYS[1] - lock name
    # ARGS[1] - token
    # return 1 if the lock was released, otherwise 0
    LUA_RELEASE_SCRIPT = """
        local token = redis.call('get', KEYS[1])
        if not token or token ~= ARGV[1] then
            return 0
        end
        redis.call('del', KEYS[1])
        return 1
    """

    # KEYS[1] - lock name
    # ARGS[1] - token
    # ARGS[2] - additional milliseconds
    # return 1 if the locks time was extended, otherwise 0
    LUA_EXTEND_SCRIPT = """
        local token = redis.call('get', KEYS[1])
        if not token or token ~= ARGV[1] then
            return 0
        end
        local expiration = redis.call('pttl', KEYS[1])
        if not expiration then
            expiration = 0
        end
        if expiration < 0 then
            return 0
        end
        redis.call('pexpire', KEYS[1], expiration + ARGV[2])
        return 1
    """

    def __init__(self, *args, **kwargs):
        super(LuaLock, self).__init__(*args, **kwargs)
        LuaLock.register_scripts(self.redis)

    @classmethod
    def register_scripts(cls, redis):
        if cls.lua_release is None:
            cls.lua_release = redis.register_script(cls.LUA_RELEASE_SCRIPT)
        if cls.lua_extend is None:
            cls.lua_extend = redis.register_script(cls.LUA_EXTEND_SCRIPT)

    async def do_release(self, expected_token):
        if not bool(
            await self.lua_release.execute(
                keys=[self.name], args=[expected_token], client=self.redis
            )
        ):
            raise LockError("Cannot release a lock that's no longer owned")

    async def do_extend(self, additional_time):
        additional_time = int(additional_time * 1000)
        if not bool(
            await self.lua_extend.execute(
                keys=[self.name],
                args=[self.local.get(), additional_time],
                client=self.redis,
            )
        ):
            raise LockError("Cannot extend a lock that's no longer owned")
        return True


class ClusterLock(LuaLock):
    """
    Cluster lock is supposed to solve lock problem in redis cluster.
    Since high availability is provided by redis cluster using master-slave model,
    the kind of lock aims to solve the fail-over problem referred in distributed lock
    post given by redis official.

    Why not use Redlock algorithm provided by official directly?
    It is impossible to make a key hashed to different nodes
    in a redis cluster and hard to generate keys
    in a specific rule and make sure they do not migrated in cluster.
    In the worst situation, all key slots may exists in one node.
    Then the availability will be the same as one key in one node.
    For more discussion please see:
    https://github.com/alisaifee/coredis/issues/55

    My solution is to take advantage of `READONLY` mode of slaves to ensure
    the lock key is synced from master to N/2 + 1 of its slaves to avoid the fail-over problem.
    Since it is a single-key solution, the migration problem also do no matter.

    Please read these article below before using this cluster lock in your app.
    https://redis.io/topics/distlock
    http://martin.kleppmann.com/2016/02/08/how-to-do-distributed-locking.html
    http://antirez.com/news/101
    """

    def __init__(self, *args, **kwargs):
        super(ClusterLock, self).__init__(*args, **kwargs)
        if not self.timeout:
            raise LockError("timeout must be provided for cluster lock")

    async def check_lock_in_slaves(self, token):
        node_manager = self.redis.connection_pool.nodes
        slot = node_manager.keyslot(self.name)
        master_node_id = node_manager.node_from_slot(slot)["node_id"]
        slave_nodes = await self.redis.cluster_slaves(master_node_id)
        count, quorum = 0, (len(slave_nodes) // 2) + 1
        conn_kwargs = self.redis.connection_pool.connection_kwargs
        conn_kwargs["readonly"] = True
        for node in slave_nodes:
            try:
                # todo: a little bit dirty here, try to reuse StrictRedis later
                # todo: it may be optimized by using a new connection pool
                conn = ClusterConnection(
                    host=node["host"], port=node["port"], **conn_kwargs
                )
                await conn.send_command("get", self.name)
                res = await conn.read_response()
                if res == token:
                    count += 1
                if count >= quorum:
                    return True
            except Exception as exc:
                warnings.warn(
                    "error {} during check lock {} status in slave nodes".format(
                        exc, self.name
                    )
                )
        return False

    async def acquire(self, blocking=None, blocking_timeout=None):
        """
        Use Redis to hold a shared, distributed lock named ``name``.
        Returns True once the lock is acquired.

        If ``blocking`` is False, always return immediately. If the lock
        was acquired, return True, otherwise return False.

        ``blocking_timeout`` specifies the maximum number of seconds to
        wait trying to acquire the lock. It should not be greater than
        expire time of the lock
        """
        sleep = self.sleep
        token = b(uuid.uuid1().hex)
        if blocking is None:
            blocking = self.blocking
        if blocking_timeout is None:
            blocking_timeout = self.blocking_timeout
        blocking_timeout = blocking_timeout or self.timeout
        stop_trying_at = mod_time.time() + min(blocking_timeout, self.timeout)

        while True:
            if await self.do_acquire(token):
                lock_acquired_at = mod_time.time()
                if await self.check_lock_in_slaves(token):
                    check_finished_at = mod_time.time()
                    # if time expends on acquiring lock is greater than given time
                    # the lock should be released manually
                    if check_finished_at > stop_trying_at:
                        await self.do_release(token)
                        return False
                    self.local.set(token)
                    # validity time is considered to be the
                    # initial validity time minus the time elapsed during check
                    await self.do_extend(lock_acquired_at - check_finished_at)
                    return True
                else:
                    await self.do_release(token)
                    return False
            if not blocking or mod_time.time() > stop_trying_at:
                return False
            if LOOP_DEPRECATED:
                await asyncio.sleep(sleep)
            else:
                await asyncio.sleep(sleep, loop=self.redis.connection_pool.loop)

    async def do_release(self, expected_token):
        await super(ClusterLock, self).do_release(expected_token)
        if await self.check_lock_in_slaves(expected_token):
            raise LockError("Lock is released in master but not in slave yet")



from __future__ import with_statement

import asyncio
import pickle
import time

import pytest

import coredis
from coredis.exceptions import ConnectionError
from coredis.utils import b


async def wait_for_message(pubsub, timeout=0.5, ignore_subscribe_messages=False):
    now = time.time()
    timeout = now + timeout
    while now < timeout:
        message = await pubsub.get_message(
            ignore_subscribe_messages=ignore_subscribe_messages, timeout=0.01
        )
        if message is not None:
            return message
        await asyncio.sleep(0.01)
        now = time.time()
    return None


def make_message(type, channel, data, pattern=None):
    return {
        "type": type,
        "pattern": pattern and pattern.encode("utf-8") or None,
        "channel": channel.encode("utf-8"),
        "data": data.encode("utf-8") if isinstance(data, str) else data,
    }


def make_subscribe_test_data(pubsub, type):
    if type == "channel":
        return {
            "p": pubsub,
            "sub_type": "subscribe",
            "unsub_type": "unsubscribe",
            "sub_func": pubsub.subscribe,
            "unsub_func": pubsub.unsubscribe,
            "keys": ["foo", "bar", "uni" + chr(56) + "code"],
        }
    elif type == "pattern":
        return {
            "p": pubsub,
            "sub_type": "psubscribe",
            "unsub_type": "punsubscribe",
            "sub_func": pubsub.psubscribe,
            "unsub_func": pubsub.punsubscribe,
            "keys": ["f*", "b*", "uni" + chr(56) + "*"],
        }
    assert False, "invalid subscribe type: %s" % type


class TestPubSubSubscribeUnsubscribe:
    async def _test_subscribe_unsubscribe(
        self, p, sub_type, unsub_type, sub_func, unsub_func, keys
    ):
        for key in keys:
            assert await sub_func(key) is None

        # should be a message for each channel/pattern we just subscribed to
        for i, key in enumerate(keys):
            assert await wait_for_message(p) == make_message(sub_type, key, i + 1)

        for key in keys:
            assert await unsub_func(key) is None

        # should be a message for each channel/pattern we just unsubscribed
        # from
        for i, key in enumerate(keys):
            i = len(keys) - 1 - i
            assert await wait_for_message(p) == make_message(unsub_type, key, i)

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_channel_subscribe_unsubscribe(self, r):
        kwargs = make_subscribe_test_data(r.pubsub(), "channel")
        await self._test_subscribe_unsubscribe(**kwargs)

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_pattern_subscribe_unsubscribe(self, r):
        kwargs = make_subscribe_test_data(r.pubsub(), "pattern")
        await self._test_subscribe_unsubscribe(**kwargs)

    async def _test_resubscribe_on_reconnection(
        self, p, sub_type, unsub_type, sub_func, unsub_func, keys
    ):

        for key in keys:
            assert await sub_func(key) is None

        # should be a message for each channel/pattern we just subscribed to
        for i, key in enumerate(keys):
            assert await wait_for_message(p) == make_message(sub_type, key, i + 1)

        # manually disconnect
        p.connection.disconnect()

        # calling get_message again reconnects and resubscribes
        # note, we may not re-subscribe to channels in exactly the same order
        # so we have to do some extra checks to make sure we got them all
        messages = []
        for i in range(len(keys)):
            messages.append(await wait_for_message(p))

        unique_channels = set()
        assert len(messages) == len(keys)
        for i, message in enumerate(messages):
            assert message["type"] == sub_type
            assert message["data"] == i + 1
            channel = message["channel"].decode("utf-8")
            unique_channels.add(channel)

        assert len(unique_channels) == len(keys)
        for channel in unique_channels:
            assert channel in keys
        await unsub_func()

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_resubscribe_to_channels_on_reconnection(self, r):
        kwargs = make_subscribe_test_data(r.pubsub(), "channel")
        await self._test_resubscribe_on_reconnection(**kwargs)

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_resubscribe_to_patterns_on_reconnection(self, r):
        kwargs = make_subscribe_test_data(r.pubsub(), "pattern")
        await self._test_resubscribe_on_reconnection(**kwargs)

    async def _test_subscribed_property(
        self, p, sub_type, unsub_type, sub_func, unsub_func, keys
    ):

        assert p.subscribed is False
        await sub_func(keys[0])
        # we're now subscribed even though we haven't processed the
        # reply from the server just yet
        assert p.subscribed is True
        assert await wait_for_message(p) == make_message(sub_type, keys[0], 1)
        # we're still subscribed
        assert p.subscribed is True

        # unsubscribe from all channels
        await unsub_func()
        # we're still technically subscribed until we process the
        # response messages from the server
        assert p.subscribed is True
        assert await wait_for_message(p) == make_message(unsub_type, keys[0], 0)
        # now we're no longer subscribed as no more messages can be delivered
        # to any channels we were listening to
        assert p.subscribed is False

        # subscribing again flips the flag back
        await sub_func(keys[0])
        assert p.subscribed is True
        assert await wait_for_message(p) == make_message(sub_type, keys[0], 1)

        # unsubscribe again
        await unsub_func()
        assert p.subscribed is True
        # subscribe to another channel before reading the unsubscribe response
        await sub_func(keys[1])
        assert p.subscribed is True
        # read the unsubscribe for key1
        assert await wait_for_message(p) == make_message(unsub_type, keys[0], 0)
        # we're still subscribed to key2, so subscribed should still be True
        assert p.subscribed is True
        # read the key2 subscribe message
        assert await wait_for_message(p) == make_message(sub_type, keys[1], 1)
        await unsub_func()
        # haven't read the message yet, so we're still subscribed
        assert p.subscribed is True
        assert await wait_for_message(p) == make_message(unsub_type, keys[1], 0)
        # now we're finally unsubscribed
        assert p.subscribed is False
        await p.unsubscribe()

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_subscribe_property_with_channels(self, r):
        kwargs = make_subscribe_test_data(r.pubsub(), "channel")
        await self._test_subscribed_property(**kwargs)

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_subscribe_property_with_patterns(self, r):
        kwargs = make_subscribe_test_data(r.pubsub(), "pattern")
        await self._test_subscribed_property(**kwargs)

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_ignore_all_subscribe_messages(self, r):
        p = r.pubsub(ignore_subscribe_messages=True)

        checks = (
            (p.subscribe, "foo"),
            (p.unsubscribe, "foo"),
            (p.psubscribe, "f*"),
            (p.punsubscribe, "f*"),
        )

        assert p.subscribed is False
        for func, channel in checks:
            assert await func(channel) is None
            assert p.subscribed is True
            assert await wait_for_message(p) is None
        assert p.subscribed is False

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_ignore_individual_subscribe_messages(self, r):
        p = r.pubsub()

        checks = (
            (p.subscribe, "foo"),
            (p.unsubscribe, "foo"),
            (p.psubscribe, "f*"),
            (p.punsubscribe, "f*"),
        )

        assert p.subscribed is False
        for func, channel in checks:
            assert await func(channel) is None
            assert p.subscribed is True
            message = await wait_for_message(p, ignore_subscribe_messages=True)
            assert message is None
        assert p.subscribed is False


class TestPubSubMessages:
    def setup_method(self, method):
        self.message = None

    def message_handler(self, message):
        self.message = message

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_published_message_to_channel(self, r):
        p = r.pubsub(ignore_subscribe_messages=True)
        await p.subscribe("foo")
        # if other tests failed, subscriber may not be cleared
        assert await r.publish("foo", "test message") >= 1

        message = await wait_for_message(p)
        assert isinstance(message, dict)
        assert message == make_message("message", "foo", "test message")
        await p.unsubscribe()

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_published_message_to_pattern(self, r):
        p = r.pubsub(ignore_subscribe_messages=True)
        await p.subscribe("foo")
        await p.psubscribe("f*")
        # 1 to pattern, 1 to channel
        assert await r.publish("foo", "test message") >= 2

        message1 = await wait_for_message(p)
        message2 = await wait_for_message(p)
        assert isinstance(message1, dict)
        assert isinstance(message2, dict)

        expected = [
            make_message("message", "foo", "test message"),
            make_message("pmessage", "foo", "test message", pattern="f*"),
        ]

        assert message1 in expected
        assert message2 in expected
        assert message1 != message2
        await p.unsubscribe("foo")

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_published_pickled_obj_to_channel(self, r):
        p = r.pubsub(ignore_subscribe_messages=True)
        await p.subscribe("foo")
        msg = pickle.dumps(Exception())
        # if other tests failed, subscriber may not be cleared
        assert await r.publish("foo", msg) >= 1

        message = await wait_for_message(p)
        assert isinstance(message, dict)
        assert message == make_message("message", "foo", msg)
        await p.unsubscribe()

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_published_pickled_obj_to_pattern(self, r):
        p = r.pubsub(ignore_subscribe_messages=True)
        await p.subscribe("foo")
        await p.psubscribe("f*")
        # 1 to pattern, 1 to channel
        msg = pickle.dumps(Exception())
        assert await r.publish("foo", msg) >= 2

        message1 = await wait_for_message(p)
        message2 = await wait_for_message(p)
        assert isinstance(message1, dict)
        assert isinstance(message2, dict)

        expected = [
            make_message("message", "foo", msg),
            make_message("pmessage", "foo", msg, pattern="f*"),
        ]

        assert message1 in expected
        assert message2 in expected
        assert message1 != message2
        await p.unsubscribe("foo")

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_channel_message_handler(self, r):
        p = r.pubsub(ignore_subscribe_messages=True)
        await p.subscribe(foo=self.message_handler)
        assert await r.publish("foo", "test message")
        assert await wait_for_message(p) is None
        assert self.message == make_message("message", "foo", "test message")
        await p.unsubscribe("foo")

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_pattern_message_handler(self, r):
        p = r.pubsub(ignore_subscribe_messages=True)
        await p.psubscribe(**{"f*": self.message_handler})
        assert await r.publish("foo", "test message")
        assert await wait_for_message(p) is None
        assert self.message == make_message(
            "pmessage", "foo", "test message", pattern="f*"
        )
        await p.unsubscribe("foo")

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_unicode_channel_message_handler(self, r):
        p = r.pubsub(ignore_subscribe_messages=True)
        channel = "uni" + chr(56) + "code"
        channels = {channel: self.message_handler}
        await p.subscribe(**channels)
        assert await r.publish(channel, "test message") == 1
        assert await wait_for_message(p) is None
        assert self.message == make_message("message", channel, "test message")
        await p.unsubscribe(channel)

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_unicode_pattern_message_handler(self, r):
        p = r.pubsub(ignore_subscribe_messages=True)
        pattern = "uni" + chr(56) + "*"
        channel = "uni" + chr(56) + "code"
        await p.psubscribe(**{pattern: self.message_handler})
        assert await r.publish(channel, "test message") == 1
        assert await wait_for_message(p) is None
        assert self.message == make_message(
            "pmessage", channel, "test message", pattern=pattern
        )
        await p.unsubscribe(channel)
        await p.punsubscribe(pattern)

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_get_message_without_subscribe(self, r):
        p = r.pubsub()
        with pytest.raises(RuntimeError) as info:
            await p.get_message()
        expect = (
            "connection not set: " "did you forget to call subscribe() or psubscribe()?"
        )
        assert expect in info.exconly()
        await p.unsubscribe()


# class TestPubSubAutoDecoding:
#     "These tests only validate that we get unicode values back"
#
#     channel = 'uni' + chr(56) + 'code'
#     pattern = 'uni' + chr(56) + '*'
#     data = 'abc' + chr(58) + '123'
#
#     def make_message(self, type, channel, data, pattern=None):
#         return {
#             'type': type,
#             'channel': channel,
#             'pattern': pattern,
#             'data': data
#         }
#
#     def setup_method(self, method):
#         self.message = None
#
#     def message_handler(self, message):
#         self.message = message
#
#     @pytest.fixture()
#     def r(self, request):
#         return _redis_client(request=request, decode_responses=True)
#
#     def test_channel_subscribe_unsubscribe(self, r):
#         p = r.pubsub()
#         p.subscribe(self.channel)
#         assert wait_for_message(p) == self.make_message('subscribe',
#                                                         self.channel, 1)
#
#         p.unsubscribe(self.channel)
#         assert wait_for_message(p) == self.make_message('unsubscribe',
#                                                         self.channel, 0)
#
#     def test_pattern_subscribe_unsubscribe(self, r):
#         p = r.pubsub()
#         p.psubscribe(self.pattern)
#         assert wait_for_message(p) == self.make_message('psubscribe',
#                                                         self.pattern, 1)
#
#         p.punsubscribe(self.pattern)
#         assert wait_for_message(p) == self.make_message('punsubscribe',
#                                                         self.pattern, 0)
#
#     def test_channel_publish(self, r):
#         p = r.pubsub(ignore_subscribe_messages=True)
#         p.subscribe(self.channel)
#         r.publish(self.channel, self.data)
#         assert wait_for_message(p) == self.make_message('message',
#                                                         self.channel,
#                                                         self.data)
#
#     def test_pattern_publish(self, r):
#         p = r.pubsub(ignore_subscribe_messages=True)
#         p.psubscribe(self.pattern)
#         r.publish(self.channel, self.data)
#         assert wait_for_message(p) == self.make_message('pmessage',
#                                                         self.channel,
#                                                         self.data,
#                                                         pattern=self.pattern)
#
#     def test_channel_message_handler(self, r):
#         p = r.pubsub(ignore_subscribe_messages=True)
#         p.subscribe(**{self.channel: self.message_handler})
#         r.publish(self.channel, self.data)
#         assert wait_for_message(p) is None
#         assert self.message == self.make_message('message', self.channel,
#                                                  self.data)
#
#         # test that we reconnected to the correct channel
#         p.connection.disconnect()
#         assert wait_for_message(p) is None  # should reconnect
#         new_data = self.data + 'new data'
#         r.publish(self.channel, new_data)
#         assert wait_for_message(p) is None
#         assert self.message == self.make_message('message', self.channel,
#                                                  new_data)
#
#     def test_pattern_message_handler(self, r):
#         p = r.pubsub(ignore_subscribe_messages=True)
#         p.psubscribe(**{self.pattern: self.message_handler})
#         r.publish(self.channel, self.data)
#         assert wait_for_message(p) is None
#         assert self.message == self.make_message('pmessage', self.channel,
#                                                  self.data,
#                                                  pattern=self.pattern)
#
#         # test that we reconnected to the correct pattern
#         p.connection.disconnect()
#         assert wait_for_message(p) is None  # should reconnect
#         new_data = self.data + 'new data'
#         r.publish(self.channel, new_data)
#         assert wait_for_message(p) is None
#         assert self.message == self.make_message('pmessage', self.channel,
#                                                  new_data,
#                                                  pattern=self.pattern)
#
#
class TestPubSubRedisDown:
    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_channel_subscribe(self, r):
        r = coredis.StrictRedis(host="localhost", port=6390)
        p = r.pubsub()
        with pytest.raises(ConnectionError):
            await p.subscribe("foo")


class TestPubSubPubSubSubcommands:
    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_pubsub_channels(self, r):
        p = r.pubsub(ignore_subscribe_messages=True)
        await p.subscribe("foo", "bar", "baz", "quux")
        channels = sorted(await r.pubsub_channels())
        assert channels == [b("bar"), b("baz"), b("foo"), b("quux")]
        await p.unsubscribe()

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_pubsub_numsub(self, r):
        p1 = r.pubsub(ignore_subscribe_messages=True)
        await p1.subscribe("foo", "bar", "baz")
        p2 = r.pubsub(ignore_subscribe_messages=True)
        await p2.subscribe("bar", "baz")
        p3 = r.pubsub(ignore_subscribe_messages=True)
        await p3.subscribe("baz")

        channels = [(b("foo"), 1), (b("bar"), 2), (b("baz"), 3)]
        assert channels == await r.pubsub_numsub("foo", "bar", "baz")
        await p1.unsubscribe()
        await p2.unsubscribe()
        await p3.unsubscribe()

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_pubsub_numpat(self, r):
        pubsub_count = await r.pubsub_numpat()
        p = r.pubsub(ignore_subscribe_messages=True)
        await p.psubscribe("*oo", "*ar", "b*z")
        assert await r.pubsub_numpat() == 3 + pubsub_count



# -*- coding: utf-8 -*-

import random

from coredis.exceptions import ConnectionError, RedisClusterException
from coredis.utils import b, hash_slot


class NodeManager:
    """
    TODO: document
    """

    RedisClusterHashSlots = 16384

    def __init__(
        self,
        startup_nodes=None,
        reinitialize_steps=None,
        skip_full_coverage_check=False,
        nodemanager_follow_cluster=False,
        **connection_kwargs,
    ):
        """
        :skip_full_coverage_check:
            Skips the check of cluster-require-full-coverage config, useful for clusters
            without the CONFIG command (like aws)
        :nodemanager_follow_cluster:
            The node manager will during initialization try the last set of nodes that
            it was operating on. This will allow the client to drift along side the cluster
            if the cluster nodes move around a slot.
        """
        self.connection_kwargs = connection_kwargs
        self.nodes = {}
        self.slots = {}
        self.startup_nodes = [] if startup_nodes is None else startup_nodes
        self.orig_startup_nodes = self.startup_nodes[:]
        self.reinitialize_counter = 0
        self.reinitialize_steps = reinitialize_steps or 25
        self._skip_full_coverage_check = skip_full_coverage_check
        self.nodemanager_follow_cluster = nodemanager_follow_cluster

        if not self.startup_nodes:
            raise RedisClusterException("No startup nodes provided")

    def encode(self, value):
        """Returns a bytestring representation of the value"""
        if isinstance(value, bytes):
            return value
        elif isinstance(value, int):
            value = b(str(value))
        elif isinstance(value, float):
            value = b(repr(value))
        elif not isinstance(value, str):
            value = str(value)
        if isinstance(value, str):
            value = value.encode()
        return value

    def keyslot(self, key):
        """Calculates keyslot for a given key"""
        key = self.encode(key)
        return hash_slot(key)

    def node_from_slot(self, slot):
        for node in self.slots[slot]:
            if node["server_type"] == "master":
                return node

    def all_nodes(self):
        for node in self.nodes.values():
            yield node

    def all_masters(self):
        for node in self.nodes.values():
            if node["server_type"] == "master":
                yield node

    def random_startup_node(self):
        return random.choice(self.startup_nodes)

    def random_startup_node_iter(self):
        """A generator that returns a random startup nodes"""
        while True:
            yield random.choice(self.startup_nodes)

    def random_node(self):
        return random.choice(list(self.nodes.values()))

    def get_redis_link(self, host, port):
        from coredis.client import StrictRedis

        allowed_keys = (
            "password",
            "stream_timeout",
            "connect_timeout",
            "retry_on_timeout",
            "ssl_context",
            "parser_class",
            "reader_read_size",
            "loop",
        )
        connection_kwargs = {
            k: v for k, v in self.connection_kwargs.items() if k in allowed_keys
        }
        return StrictRedis(
            host=host, port=port, decode_responses=True, **connection_kwargs
        )

    async def initialize(self):
        """
        Initializes the slots cache by asking all startup nodes what the
        current cluster configuration is.

        TODO: Currently the last node will have the last say about how the configuration is setup.
        Maybe it should stop to try after it have correctly covered all slots or when one node is
        reached and it could execute CLUSTER SLOTS command.
        """
        nodes_cache = {}
        tmp_slots = {}

        all_slots_covered = False
        disagreements = []
        startup_nodes_reachable = False

        nodes = self.orig_startup_nodes

        # With this option the client will attempt to connect to any of the previous set of nodes
        # instead of the original set of nodes
        if self.nodemanager_follow_cluster:
            nodes = self.startup_nodes

        for node in nodes:
            try:
                r = self.get_redis_link(host=node["host"], port=node["port"])
                cluster_slots = await r.cluster_slots()
                startup_nodes_reachable = True
            except ConnectionError:
                continue
            except Exception:
                raise RedisClusterException(
                    f'ERROR sending "cluster slots" command to redis server: {node}'
                )

            all_slots_covered = True

            # If there's only one server in the cluster, its ``host`` is ''
            # Fix it to the host in startup_nodes
            if len(cluster_slots) == 1 and len(self.startup_nodes) == 1:
                single_node_slots = cluster_slots.get(
                    (0, self.RedisClusterHashSlots - 1)
                )[0]
                if len(single_node_slots["host"]) == 0:
                    single_node_slots["host"] = self.startup_nodes[0]["host"]
                    single_node_slots["server_type"] = "master"

            # No need to decode response because StrictRedis should handle that for us...
            for min_slot, max_slot in cluster_slots:
                nodes = cluster_slots.get((min_slot, max_slot))
                master_node, slave_nodes = nodes[0], nodes[1:]

                if master_node["host"] == "":
                    master_node["host"] = node["host"]
                self.set_node_name(master_node)
                nodes_cache[master_node["name"]] = master_node

                for i in range(min_slot, max_slot + 1):
                    if i not in tmp_slots:
                        tmp_slots[i] = [master_node]

                        for slave_node in slave_nodes:
                            self.set_node_name(slave_node)
                            nodes_cache[slave_node["name"]] = slave_node
                            tmp_slots[i].append(slave_node)
                    else:
                        # Validate that 2 nodes want to use the same slot cache setup
                        if tmp_slots[i][0]["name"] != node["name"]:
                            disagreements.append(
                                "{0} vs {1} on slot: {2}".format(
                                    tmp_slots[i][0]["name"], node["name"], i
                                ),
                            )

                            if len(disagreements) > 5:
                                raise RedisClusterException(
                                    (
                                        "startup_nodes could not agree on a valid slots cache."
                                        f" {', '.join(disagreements)}"
                                    )
                                )

                self.populate_startup_nodes()
                self.refresh_table_asap = False

            if self._skip_full_coverage_check:
                need_full_slots_coverage = False
            else:
                need_full_slots_coverage = await self.cluster_require_full_coverage(
                    nodes_cache
                )

            # Validate if all slots are covered or if we should try next startup node
            for i in range(0, self.RedisClusterHashSlots):
                if i not in tmp_slots and need_full_slots_coverage:
                    all_slots_covered = False

            if all_slots_covered:
                # All slots are covered and application can continue to execute
                break

        if not startup_nodes_reachable:
            raise RedisClusterException(
                "Redis Cluster cannot be connected. "
                "Please provide at least one reachable node."
            )

        if not all_slots_covered:
            raise RedisClusterException(
                "Not all slots are covered after query all startup_nodes. "
                "{0} of {1} covered...".format(
                    len(tmp_slots), self.RedisClusterHashSlots
                )
            )

        # Set the tmp variables to the real variables
        self.slots = tmp_slots
        self.nodes = nodes_cache
        self.reinitialize_counter = 0

    async def increment_reinitialize_counter(self, ct=1):
        for i in range(1, ct):
            self.reinitialize_counter += 1
            if self.reinitialize_counter % self.reinitialize_steps == 0:
                await self.initialize()

    async def cluster_require_full_coverage(self, nodes_cache):
        """
        If exists 'cluster-require-full-coverage no' config on redis servers,
        then even all slots are not covered, cluster still will be able to
        respond
        """
        nodes = nodes_cache or self.nodes

        async def node_require_full_coverage(node):
            r_node = self.get_redis_link(host=node["host"], port=node["port"])
            node_config = await r_node.config_get("cluster-require-full-coverage")
            return "yes" in node_config.values()

        # at least one node should have cluster-require-full-coverage yes
        for node in nodes.values():
            if await node_require_full_coverage(node):
                return True
        return False

    def set_node_name(self, n):
        """
        Formats the name for the given node object

        # TODO: This shold not be constructed this way. It should update the name of the node in
        the node cache dict
        """
        if "name" not in n:
            n["name"] = "{0}:{1}".format(n["host"], n["port"])

    def set_node(self, host, port, server_type=None):
        """Updates data for a node"""
        node_name = "{0}:{1}".format(host, port)
        node = {
            "host": host,
            "port": port,
            "name": node_name,
            "server_type": server_type,
        }
        self.nodes[node_name] = node
        return node

    def populate_startup_nodes(self):
        """
        Do something with all startup nodes and filters out any duplicates
        """
        for item in self.startup_nodes:
            self.set_node_name(item)

        for n in self.nodes.values():
            if n not in self.startup_nodes:
                self.startup_nodes.append(n)

        # freeze it so we can set() it
        uniq = {frozenset(node.items()) for node in self.startup_nodes}
        # then thaw it back out into a list of dicts
        self.startup_nodes = [dict(node) for node in uniq]

    async def reset(self):
        """Drops all node data and start over from startup_nodes"""
        await self.initialize()



import warnings

from coredis.utils import NodeFlag, bool_ok, dict_merge, list_keys_to_dict, nativestr

SENTINEL_STATE_TYPES = {
    "can-failover-its-master": int,
    "config-epoch": int,
    "down-after-milliseconds": int,
    "failover-timeout": int,
    "info-refresh": int,
    "last-hello-message": int,
    "last-ok-ping-reply": int,
    "last-ping-reply": int,
    "last-ping-sent": int,
    "master-link-down-time": int,
    "master-port": int,
    "num-other-sentinels": int,
    "num-slaves": int,
    "o-down-time": int,
    "pending-commands": int,
    "parallel-syncs": int,
    "port": int,
    "quorum": int,
    "role-reported-time": int,
    "s-down-time": int,
    "slave-priority": int,
    "slave-repl-offset": int,
    "voted-leader-epoch": int,
}


def pairs_to_dict_typed(response, type_info):
    it = iter(response)
    result = {}
    for key, value in zip(it, it):
        if key in type_info:
            try:
                value = type_info[key](value)
            except Exception:
                # if for some reason the value can't be coerced, just use
                # the string value
                pass
        result[key] = value
    return result


def parse_sentinel_state(item):
    result = pairs_to_dict_typed(item, SENTINEL_STATE_TYPES)
    flags = set(result["flags"].split(","))
    for name, flag in (
        ("is_master", "master"),
        ("is_slave", "slave"),
        ("is_sdown", "s_down"),
        ("is_odown", "o_down"),
        ("is_sentinel", "sentinel"),
        ("is_disconnected", "disconnected"),
        ("is_master_down", "master_down"),
    ):
        result[name] = flag in flags
    return result


def parse_sentinel_master(response):
    return parse_sentinel_state(map(nativestr, response))


def parse_sentinel_masters(response):
    result = {}
    for item in response:
        state = parse_sentinel_state(map(nativestr, item))
        result[state["name"]] = state
    return result


def parse_sentinel_slaves_and_sentinels(response):
    return [parse_sentinel_state(map(nativestr, item)) for item in response]


def parse_sentinel_get_master(response):
    return response and (response[0], int(response[1])) or None


class SentinelCommandMixin:
    RESPONSE_CALLBACKS = {
        "SENTINEL GET-MASTER-ADDR-BY-NAME": parse_sentinel_get_master,
        "SENTINEL MASTER": parse_sentinel_master,
        "SENTINEL MASTERS": parse_sentinel_masters,
        "SENTINEL MONITOR": bool_ok,
        "SENTINEL REMOVE": bool_ok,
        "SENTINEL SENTINELS": parse_sentinel_slaves_and_sentinels,
        "SENTINEL SET": bool_ok,
        "SENTINEL SLAVES": parse_sentinel_slaves_and_sentinels,
    }

    async def sentinel(self, *args):
        """Redis Sentinel's SENTINEL command."""
        warnings.warn(DeprecationWarning("Use the individual sentinel_* methods"))

    async def sentinel_get_master_addr_by_name(self, service_name):
        """Returns a (host, port) pair for the given ``service_name``"""
        return await self.execute_command(
            "SENTINEL GET-MASTER-ADDR-BY-NAME", service_name
        )

    async def sentinel_master(self, service_name):
        """Returns a dictionary containing the specified masters state."""
        return await self.execute_command("SENTINEL MASTER", service_name)

    async def sentinel_masters(self):
        """Returns a list of dictionaries containing each master's state."""
        return await self.execute_command("SENTINEL MASTERS")

    async def sentinel_monitor(self, name, ip, port, quorum):
        """Adds a new master to Sentinel to be monitored"""
        return await self.execute_command("SENTINEL MONITOR", name, ip, port, quorum)

    async def sentinel_remove(self, name):
        """Removes a master from Sentinel's monitoring"""
        return await self.execute_command("SENTINEL REMOVE", name)

    async def sentinel_sentinels(self, service_name):
        """Returns a list of sentinels for ``service_name``"""
        return await self.execute_command("SENTINEL SENTINELS", service_name)

    async def sentinel_set(self, name, option, value):
        """Sets Sentinel monitoring parameters for a given master"""
        return await self.execute_command("SENTINEL SET", name, option, value)

    async def sentinel_slaves(self, service_name):
        """Returns a list of slaves for ``service_name``"""
        return await self.execute_command("SENTINEL SLAVES", service_name)


class ClusterSentinelCommands(SentinelCommandMixin):
    NODES_FLAGS = dict_merge(
        list_keys_to_dict(
            [
                "SENTINEL GET-MASTER-ADDR-BY-NAME",
                "SENTINEL MASTER",
                "SENTINEL MASTERS",
                "SENTINEL MONITOR",
                "SENTINEL REMOVE",
                "SENTINEL SENTINELS",
                "SENTINEL SET",
                "SENTINEL SLAVES",
            ],
            NodeFlag.BLOCKED,
        )
    )



#!/usr/bin/env python
import pathlib
import re
import sys
import versioneer

__author__ = "Ali-Akber Saifee"
__email__ = "ali@indydevs.org"
__copyright__ = "Copyright 2022, Ali-Akber Saifee"

try:
    from setuptools import setup
    from setuptools.command.test import test as TestCommand
    from setuptools.command.build_ext import build_ext
    from setuptools.extension import Extension

    class PyTest(TestCommand):
        def finalize_options(self):
            TestCommand.finalize_options(self)
            self.test_args = []
            self.test_suite = True

        def run_tests(self):
            # import here, because outside the eggs aren't loaded
            import pytest

            errno = pytest.main(self.test_args)
            sys.exit(errno)


except ImportError:

    from distutils.core import setup, Extension
    from distutils.command.build_ext import build_ext

    def PyTest(x):
        x


class custom_build_ext(build_ext):
    """
    NOTE: This code was originally taken from tornado.

    Allows C extension building to fail.

    The C extension speeds up crc16, but is not essential.
    """

    warning_message = """
********************************************************************
{target} could not
be compiled. No C extensions are essential for coredis to run,
although they do result in significant speed improvements for
websockets.
{comment}

Here are some hints for popular operating systems:

If you are seeing this message on Linux you probably need to
install GCC and/or the Python development package for your
version of Python.

Debian and Ubuntu users should issue the following command:

    $ sudo apt-get install build-essential python-dev

RedHat and CentOS users should issue the following command:

    $ sudo yum install gcc python-devel

Fedora users should issue the following command:

    $ sudo dnf install gcc python-devel

If you are seeing this message on OSX please read the documentation
here:

https://api.mongodb.org/python/current/installation.html#osx
********************************************************************
"""

    def run(self):
        try:
            super().run()
        except Exception as e:
            self.warn(e)
            self.warn(
                self.warning_message.format(
                    target="Extension modules",
                    comment=(
                        "There is an issue with your platform configuration "
                        "- see above."
                    ),
                )
            )

    def build_extension(self, ext):
        try:
            super().build_extension(ext)
        except Exception as e:
            self.warn(e)
            self.warn(
                self.warning_message.format(
                    target="The {} extension ".format(ext.name),
                    comment=(
                        "The output above this warning shows how the "
                        "compilation failed."
                    ),
                )
            )


_ROOT_DIR = pathlib.Path(__file__).parent

with open(str(_ROOT_DIR / "README.rst")) as f:
    long_description = f.read()

setup(
    name="coredis",
    version=versioneer.get_version(),
    description="Python async client for Redis key-value store",
    long_description=long_description,
    url="https://github.com/alisaifee/coredis",
    author=__author__,
    author_email=__email__,
    maintainer=__author__,
    maintainer_email=__email__,
    keywords=["Redis", "key-value store", "asyncio"],
    license="MIT",
    packages=["coredis", "coredis.commands"],
    python_requires=">=3.7",
    extras_require={"hiredis": ["hiredis>=0.2.0"]},
    tests_require=["pytest", "pytest_asyncio>=0.5.0"],
    cmdclass=versioneer.get_cmdclass({"test": PyTest, "build_ext": custom_build_ext}),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    ext_modules=[Extension(name="coredis.speedups", sources=["coredis/speedups.c"]),],
)
